import numpy as np
import shapely

from threedigrid_builder.constants import ContentType
from threedigrid_builder.grid.zero_d import (
    ImperviousSurfaces,
    SURFACE_CLASS_MAP,
    SURFACE_INCLINATION_MAP,
    SURFACE_TYPE_PROPERTIES,
    SurfaceParams,
    Surfaces,
)


def test_surface_params():
    """
    Check surface params conversions
    """
    surface_class = []
    surface_inclination = []

    expected = {}
    for surface_cls in SURFACE_CLASS_MAP:
        for surface_incl in SURFACE_INCLINATION_MAP:
            surface_class.append(surface_cls)
            surface_inclination.append(surface_incl)
            values_dict = SURFACE_TYPE_PROPERTIES[
                f"{SURFACE_CLASS_MAP[surface_cls]}_{SURFACE_INCLINATION_MAP[surface_incl]}"
            ]
            for k, v in values_dict.items():
                if k not in expected:
                    expected[k] = []
                expected[k].append(v)

    params = SurfaceParams.from_surface_class_and_inclination(
        np.array(surface_class), np.array(surface_inclination)
    )

    for k, v in expected.items():
        value_to_check = getattr(params, k)
        assert np.all(value_to_check == np.array(v, dtype=value_to_check.dtype))


def test_surfaces(grid_all):
    surface = Surfaces(
        id=[1, 2, 3],
        function=[b"1", b"2", b"3"],
        outflow_delay=[1.0, 2.0, 2.0],
        surface_layer_thickness=[1.0, 2.0, 2.0],
        infiltration=[0.0, 1.0, 1.0],
        max_infiltration_capacity=[1.0, 2.0, 2.0],
        min_infiltration_capacity=[0.0, 1.0, 1.0],
        infiltration_decay_constant=[1.1, 1.2, 1.2],
        infiltration_recovery_constant=[0.9, 1.2, 1.2],
        surface_id=[1, 2, 2],
        code=[b"1", b"2", b"2"],
        display_name=[b"d1", b"d2", b"d2"],
        nr_of_inhabitants=[1000.0, 2000.0, 2000.0],
        area=[100.0, 200.0, 200.0],
        dry_weather_flow=[1.0, 2.0, 2.0],
        the_geom=[shapely.points([1.0, 2.0]), None, None],
        connection_node_id=[1, 2, 1],
        connection_node_the_geom=[
            shapely.points([1.0, 2.0]),
            shapely.points([2.0, 3.0]),
            shapely.points([1.0, 2.0]),
        ],
        percentage=[100.0, 50.0, 50.0],
    )

    grid_all.nodes.content_type[1:] = ContentType.TYPE_V2_CONNECTION_NODES.value
    grid_surfaces = surface.as_grid_surfaces()

    expected = (
        ("id", (2,), "int32"),
        ("code", (2,), "object"),
        ("display_name", (2,), "object"),
        ("function", (2,), "object"),
        ("area", (2,), "float64"),
        ("centroid_x", (2,), "float64"),
        ("centroid_y", (2,), "float64"),
        ("dry_weather_flow", (2,), "float64"),
        ("nr_of_inhabitants", (2,), "float64"),
        ("infiltration_flag", (2,), "bool"),
        ("outflow_delay", (2,), "float64"),
        ("storage_limit", (2,), "float64"),
        ("fb", (2,), "float64"),
        ("fe", (2,), "float64"),
        ("ka", (2,), "float64"),
        ("kh", (2,), "float64"),
        ("surface_class", (2,), "object"),
        ("surface_inclination", (2,), "object"),
        ("surface_sub_class", (2,), "object"),
    )

    for name, shape, dtype in expected:
        value = getattr(grid_surfaces, name)
        assert value.shape == shape
        assert value.dtype == np.dtype(dtype)

    # Second surface centroid is based on geometries of both coupled connection_node geoms
    assert np.all(grid_surfaces.centroid_x == np.array([1.0, 1.5]))
    assert np.all(grid_surfaces.centroid_y == np.array([2.0, 2.5]))

    # Check surface maps
    expected = (
        ("fac", (3,), "float64"),
        ("imp", (3,), "int32"),
        ("nxc", (3,), "float64"),
        ("nyc", (3,), "float64"),
        ("pk", (3,), "int32"),
        ("cci", (3,), "int32"),
    )

    grid_surface_maps = surface.as_surface_maps(grid_all.nodes, grid_all.nodes_embedded)

    for name, shape, dtype in expected:
        value = getattr(grid_surface_maps, name)
        assert value.shape == shape
        assert value.dtype == np.dtype(dtype)


def test_impervioussurfaces(grid_all):
    surface = ImperviousSurfaces(
        id=[1, 2, 3],
        surface_inclination=["hellend", "vlak", "vlak"],
        surface_class=["pand", "onverhard", "onverhard"],
        surface_sub_class=[b"1", b"2", b"3"],
        surface_id=[1, 2, 2],
        code=[b"1", b"2", b"2"],
        display_name=[b"d1", b"d2", b"d2"],
        nr_of_inhabitants=[1000.0, 2000.0, 2000.0],
        area=[100.0, 200.0, 200.0],
        dry_weather_flow=[1.0, 2.0, 2.0],
        the_geom=[shapely.points([1.0, 2.0]), None, None],
        connection_node_id=[1, 2, 1],
        connection_node_the_geom=[
            shapely.points([1.0, 2.0]),
            shapely.points([2.0, 3.0]),
            shapely.points([1.0, 2.0]),
        ],
        percentage=[100.0, 50.0, 50.0],
    )

    grid_all.nodes.content_type[1:] = ContentType.TYPE_V2_CONNECTION_NODES.value
    grid_surfaces = surface.as_grid_surfaces()

    expected = (
        ("id", (2,), "int32"),
        ("code", (2,), "object"),
        ("display_name", (2,), "object"),
        ("function", (2,), "object"),
        ("area", (2,), "float64"),
        ("centroid_x", (2,), "float64"),
        ("centroid_y", (2,), "float64"),
        ("dry_weather_flow", (2,), "float64"),
        ("nr_of_inhabitants", (2,), "float64"),
        ("infiltration_flag", (2,), "bool"),
        ("outflow_delay", (2,), "float64"),
        ("storage_limit", (2,), "float64"),
        ("fb", (2,), "float64"),
        ("fe", (2,), "float64"),
        ("ka", (2,), "float64"),
        ("kh", (2,), "float64"),
        ("surface_class", (2,), "object"),
        ("surface_inclination", (2,), "object"),
        ("surface_sub_class", (2,), "object"),
    )

    for name, shape, dtype in expected:
        value = getattr(grid_surfaces, name)
        assert value.shape == shape
        assert value.dtype == np.dtype(dtype)

    # Second surface centroid is based on geometries of both coupled connection_node geoms
    assert np.all(grid_surfaces.centroid_x == np.array([1.0, 1.5]))
    assert np.all(grid_surfaces.centroid_y == np.array([2.0, 2.5]))

    # Check one of the SurfaceParams
    assert np.all(
        grid_surfaces.outflow_delay
        == np.array([0.5 / 60.0, 0.2 / 60.0], dtype="float32")
    )

    expected = (
        ("fac", (3,), "float64"),
        ("imp", (3,), "int32"),
        ("nxc", (3,), "float64"),
        ("nyc", (3,), "float64"),
        ("pk", (3,), "int32"),
        ("cci", (3,), "int32"),
    )
    grid_surface_maps = surface.as_surface_maps(grid_all.nodes, grid_all.nodes_embedded)

    for name, shape, dtype in expected:
        value = getattr(grid_surface_maps, name)
        assert value.shape == shape
        assert value.dtype == np.dtype(dtype)



import itertools
from unittest import mock

import numpy as np
import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_array_equal

from threedigrid_builder.base import Nodes
from threedigrid_builder.grid import ConnectionNodes, Pipes


@pytest.fixture
def connection_nodes():
    # Used to map connection_node_start/end_id to an index (sequence id)
    return ConnectionNodes(
        id=[21, 25, 33, 42],
        the_geom=shapely.points([(0, 21), (1, 25), (2, 33), (3, 42)]),
        drain_level=[0.0, 20.0, 30.0, 10.0],
    )


@pytest.fixture
def pipes():
    return Pipes(
        id=[1, 2],
        dist_calc_points=[5.0, np.nan],
        code=["one", "two"],
        connection_node_start_id=[21, 21],
        connection_node_end_id=[25, 42],
        calculation_type=[2, 1],
        invert_level_start_point=[3.0, 5.0],
        invert_level_end_point=[4.0, 6.0],
    )


@pytest.mark.parametrize(
    "pipe_ids,ds,expected",
    [
        ([1], [5.0], [3.5]),
        ([2], [0.5], [5.5]),
        ([1, 2], [5.0, 0.5], [3.5, 5.5]),
        ([2, 1], [0.5, 5.0], [5.5, 3.5]),
        ([1, 1, 2, 2], [5.0, 7.5, 0.5, 0.25], [3.5, 3.75, 5.5, 5.25]),
    ],
)
def test_compute_bottom_level(pipe_ids, ds, pipes, expected):
    # set geometries with lengths 10 and 1 (resp. id 1 and 2)
    # invert levels are [3, 4] for id=1 and [5, 6] for id=2
    pipes.the_geom = shapely.linestrings([[(0, 0), (0, 10)], [(2, 2), (3, 2)]])

    actual = pipes.compute_bottom_level(pipe_ids, ds)

    assert_almost_equal(actual, expected)


def test_compute_bottom_level_raises_no_geom(pipes):
    with pytest.raises(ValueError, match=".*Call set_geometries first.*"):
        pipes.compute_bottom_level([1], [5.0])


@pytest.mark.parametrize(
    "pipe_ids,ds",
    [
        ([1], [10.1]),
        ([2], [-1e-7]),
    ],
)
def test_compute_bottom_level_raises_out_of_bounds(pipe_ids, ds, pipes):
    pipes.the_geom = shapely.linestrings([[(0, 0), (0, 10)], [(2, 2), (3, 2)]])
    with pytest.raises(ValueError, match=".*outside of the linear object bounds.*"):
        pipes.compute_bottom_level(pipe_ids, ds)


def test_get_1d2d_exchange_levels(pipes):
    content_pk = np.array([1, 1, 2])
    s1d = np.array([12.0, 13.0, 15.0])
    connection_nodes = mock.Mock()

    with mock.patch.object(pipes, "compute_drain_level") as compute_drain_level:
        compute_drain_level.return_value = np.array([1.8, 2.5, 3.2])
        actual = pipes.get_1d2d_exchange_levels(
            content_pk=content_pk, s1d=s1d, connection_nodes=connection_nodes
        )

        _, kwargs = compute_drain_level.call_args
        assert_array_equal(kwargs["ids"], content_pk)
        assert_array_equal(kwargs["s"], s1d)
        assert kwargs["connection_nodes"] is connection_nodes

    # interpolation between manhole drain levels is further tested elsewhere
    assert_array_equal(actual, compute_drain_level.return_value)


def test_is_closed(pipes):
    actual = pipes.is_closed(np.array([1, 3]))

    assert_array_equal(actual, [True, True])


@pytest.mark.parametrize(
    "thickness,hc_out,hc_in,expected",
    [
        (0.1, 3.0, 2.0, True),
        (0.0, 3.0, 2.0, False),
        (np.nan, 3.0, 2.0, False),
        (0.1, np.nan, 2.0, False),
        (0.1, 3.0, np.nan, False),
    ],
)
def test_has_groundwater_exchange(thickness, hc_out, hc_in, expected):
    pipes = Pipes(
        id=[1],
        exchange_thickness=thickness,
        hydraulic_conductivity_out=hc_out,
        hydraulic_conductivity_in=hc_in,
    )

    actual = pipes.has_groundwater_exchange

    assert len(actual) == 1
    assert actual[0] == expected


@pytest.mark.parametrize(
    "thickness,hc_out,hc_in,res_out,res_in",
    [
        (0.1, 3.0, 2.0, 30.0, 20.0),
        (np.nan, 3.0, 2.0, np.nan, np.nan),
        (0.1, np.nan, 2.0, np.nan, 20.0),
        (0.1, 3.0, np.nan, 30.0, np.nan),
    ],
)
def test_set_hydraulic_resistance(thickness, hc_out, hc_in, res_in, res_out):
    pipes = Pipes(
        the_geom=shapely.linestrings([[(0, 0), (6, 0), (6, 6)]]),
        dist_calc_points=np.array([5.0]),
        id=np.array([1]),
        code=np.array(["one"]),
        connection_node_start_id=np.array([21]),
        connection_node_end_id=np.array([42]),
        calculation_type=np.array([2]),
    )
    pipes.exchange_thickness = np.array([thickness])
    pipes.hydraulic_conductivity_out = np.array([hc_out])
    pipes.hydraulic_conductivity_in = np.array([hc_in])

    cn = ConnectionNodes(id=[21, 42], the_geom=shapely.points([(0, 21), (1, 25)]))
    nodes = Nodes(id=[10], content_pk=[1], s1d=[0.5])
    lines = pipes.get_lines(cn, nodes, itertools.count(start=0))

    assert_array_equal(lines.hydraulic_resistance_in, [res_in, res_in])
    assert_array_equal(lines.hydraulic_resistance_out, [res_out, res_out])



import itertools
from typing import Iterator, Tuple

import numpy as np
import shapely

from threedigrid_builder.base import Array, Lines, Nodes
from threedigrid_builder.constants import (
    BoundaryType,
    CalculationType,
    ContentType,
    LineType,
    NodeType,
)
from threedigrid_builder.exceptions import SchematisationError

__all__ = ["BoundaryConditions1D", "BoundaryConditions2D"]


GROUNDWATER_BOUNDARY_TYPES = frozenset(
    {BoundaryType.GROUNDWATERLEVEL, BoundaryType.GROUNDWATERDISCHARGE}
)


class BoundaryCondition1D:
    id: int
    boundary_type: BoundaryType
    connection_node_id: int


class BoundaryConditions1D(Array[BoundaryCondition1D]):
    def apply(self, grid):
        """Apply 1D boundary conditions to a Grid

        Fields on nodes/lines that have a boundary condition be adjusted:
        - nodes.calculation_type: set (overridden) to BOUNDARY_NODE
        - nodes.boundary_id: from BoundaryConditions1D.id
        - nodes.boundary_type: from BoundaryConditions1D.boundary_type
        - nodes.node_type: set to NODE_1D_BOUNDARIES
        - lines.boundary_id: from BoundaryConditions1D.id

        Note:
        - it is assumed that nodes.content_pk is sorted for connection nodes
        - this method should go before 1D-2D connections and node embedding
        """
        # lookup corresponding nodes
        cn_idx = np.where(
            grid.nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES
        )[0]
        idx = cn_idx[
            np.searchsorted(grid.nodes.content_pk[cn_idx], self.connection_node_id)
        ]
        if not np.all(grid.nodes.content_pk[idx] == self.connection_node_id):
            missing = np.where(grid.nodes.content_pk[idx] != self.connection_node_id)[0]
            raise SchematisationError(
                f"1D boundary conditions {self.id[missing].tolist()} refer to missing "
                f"connection nodes {self.connection_node_id[missing].tolist()}."
            )

        # lookup the lines connecting to these nodes
        ids = grid.nodes.index_to_id(idx)
        line_idx, start_or_end = np.where(np.isin(grid.lines.line, ids))
        # lookup the boundary condition node and other node corresponding to these lines
        bc_node_id = grid.lines.line[line_idx, start_or_end]
        # this trick transforms 0 to 1 and 1 to 0:
        start_or_end_inv = (~(start_or_end.astype(bool))).astype(int)
        other_node_id = grid.lines.line[line_idx, start_or_end_inv]

        if not np.isin(ids, bc_node_id).all():
            missing = ~np.isin(ids, bc_node_id)
            raise SchematisationError(
                f"1D boundary conditions {self.id[missing].tolist()} refer to connection nodes "
                f"that have no attached objects ({self.connection_node_id[missing].tolist()})."
            )
        if len(ids) != len(bc_node_id):
            sorted_cn_id, counts = np.unique(bc_node_id, return_counts=True)
            duplicate_cn_id = sorted_cn_id[counts > 1]
            raise SchematisationError(
                f"Only one attached object is allowed for connection nodes "
                f"{duplicate_cn_id.tolist()} with a 1D boundary condition."
            )
        if np.isin(ids, other_node_id).any():
            both_sides = np.isin(ids, other_node_id)
            raise SchematisationError(
                f"1D boundary conditions {self.id[both_sides].tolist()} are too close to other "
                f"boundary conditions."
            )

        # set node attributes
        grid.nodes.calculation_type[idx] = CalculationType.BOUNDARY_NODE
        grid.nodes.boundary_id[idx] = self.id
        grid.nodes.boundary_type[idx] = self.boundary_type
        grid.nodes.node_type[idx] = NodeType.NODE_1D_BOUNDARIES
        # set line attributes
        grid.lines.is_1d_boundary[line_idx] = 1


class BoundaryCondition2D:
    id: int
    boundary_type: BoundaryType
    the_geom: shapely.Geometry


class BoundaryConditions2D(Array[BoundaryCondition2D]):
    def get_intersecting_node_idx(
        self, idx: int, cell_tree: shapely.STRtree
    ) -> np.ndarray:
        bc_geom = self.the_geom[idx]

        x1, y1, x2, y2 = shapely.bounds(bc_geom)
        is_horizontal = (x2 - x1) > (y2 - y1)

        node_idx = np.sort(cell_tree.query(bc_geom))
        if node_idx.size == 0:
            # Pragmatic fix; because of reprojection issues, the boundary condition
            # can be right outside the cells.

            # Start by getting the closest point inside the model area (all cells)
            nearest_cell = cell_tree.geometries[cell_tree.nearest(bc_geom)]
            shortest_line = shapely.shortest_line(nearest_cell, bc_geom)
            x1, _, x2, _ = shapely.bounds(nearest_cell)
            if shapely.length(shortest_line) < (x2 - x1):
                cell_coords, _ = shapely.get_coordinates(shortest_line)

                # Adjust the x / y coordinate to the x / y coordinate of the point
                bc_coords = shapely.get_coordinates(bc_geom)
                bc_coords[:, int(is_horizontal)] = cell_coords[int(is_horizontal)]
                bc_geom = shapely.set_coordinates(bc_geom, bc_coords)

                # Query again for intersecting cells
                node_idx = np.sort(cell_tree.query(bc_geom))

        return node_idx, is_horizontal

    def get_neighoring_node_idx(
        self,
        idx: int,
        nodes: Nodes,
        node_idx: np.ndarray,
        quad_idx: np.ndarray,
        is_horizontal: bool,
    ) -> np.ndarray:
        nodk = nodes.nodk[node_idx]
        sz = 2 ** (nodk - 1)
        # nodm and nodn are 1-based indices into quad_idx. So we subtract 1 from it.
        nodm = nodes.nodm[node_idx] - 1
        nodn = nodes.nodn[node_idx] - 1
        # Resized nodm, nodn (called x, y) so that they index into quad_idx.
        # The +1 accounts for the padding of quad_idx.
        x = nodm * sz + 1
        y = nodn * sz + 1
        # just to be sure, can be removed later:
        assert np.all(quad_idx[x, y] == (node_idx + 1))

        if is_horizontal:  # horizontal BC, looking at vertical neigbours
            # A cell has no bottom (upstream) neighbor if:
            # - the nodgrid entry before (under) at the left & right sides is 0
            before = (quad_idx[x, y - 1] == 0) & (quad_idx[x + sz - 1, y - 1] == 0)
            # A cell has no top (downstream) neighbor if:
            # - the nodgrid entry after (above) at the left & right sides is 0
            after = (quad_idx[x, y + sz] == 0) & (quad_idx[x + sz - 1, y + sz] == 0)
        else:  # vertical BC, looking at horizontal neigbours
            # A cell has no left (upstream) neighbor if:
            # - the nodgrid entry before (to the left) at bottom & top sides are 0
            before = (quad_idx[x - 1, y] == 0) & (quad_idx[x - 1, y + sz - 1] == 0)
            # A cell has no right (downstream) neighbor if:
            # - the nodgrid entry after (to the right) at bottom & top sides is 0
            after = (quad_idx[x + sz, y] == 0) & (quad_idx[x + sz, y + sz - 1] == 0)

        # The edge of the boundary is whichever side has the most edges
        n_edges_before = np.count_nonzero(before)
        n_edges_after = np.count_nonzero(after)
        if n_edges_after == 0 and n_edges_before == 0:
            raise SchematisationError(
                f"2D boundary condition {self.id[idx]} does not touch any edge "
                f"cell."
            )
        elif n_edges_after == n_edges_before:
            s = "top and bottom" if is_horizontal else "right and left"
            raise SchematisationError(
                f"2D boundary condition {self.id[idx]} touches cells that have "
                f"equal numbers of {s} edges."
            )

        is_before = n_edges_before > n_edges_after  # is_upstream

        return node_idx[before if is_before else after], is_before

    def adapt_node_idx_groundwater(self, idx: int, nodes: Nodes, node_idx):
        is_groundwater = self.boundary_type_is_groundwater(self.boundary_type[idx])
        if is_groundwater:
            n_groundwater_cells = nodes.n_groundwater_cells
            if n_groundwater_cells == 0:
                raise SchematisationError(
                    f"2D boundary condition {self.id[idx]} has groundwater type while "
                    f"there are no groundwater cells."
                )

            node_idx += n_groundwater_cells
        return is_groundwater

    @staticmethod
    def get_kcu(is_horizontal: bool, is_before: bool, is_groundwater: bool) -> LineType:
        return {
            (True, True, False): LineType.LINE_2D_BOUNDARY_SOUTH,
            (True, False, False): LineType.LINE_2D_BOUNDARY_NORTH,
            (False, True, False): LineType.LINE_2D_BOUNDARY_WEST,
            (False, False, False): LineType.LINE_2D_BOUNDARY_EAST,
            (True, True, True): LineType.LINE_2D_GROUNDWATER_BOUNDARY_SOUTH,
            (True, False, True): LineType.LINE_2D_GROUNDWATER_BOUNDARY_NORTH,
            (False, True, True): LineType.LINE_2D_GROUNDWATER_BOUNDARY_WEST,
            (False, False, True): LineType.LINE_2D_GROUNDWATER_BOUNDARY_EAST,
        }[(is_horizontal, is_before, is_groundwater)]

    @staticmethod
    def get_edge_coord_col(kcu: LineType) -> int:
        return {
            LineType.LINE_2D_BOUNDARY_SOUTH: 1,
            LineType.LINE_2D_BOUNDARY_NORTH: 3,
            LineType.LINE_2D_BOUNDARY_WEST: 0,
            LineType.LINE_2D_BOUNDARY_EAST: 2,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_SOUTH: 1,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_NORTH: 3,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_WEST: 0,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_EAST: 2,
        }[kcu]

    @staticmethod
    def get_cross_pix_coords_cols(kcu: LineType) -> Tuple[int, int, int, int]:
        return {
            LineType.LINE_2D_BOUNDARY_SOUTH: (0, 1, 2, 1),
            LineType.LINE_2D_BOUNDARY_NORTH: (0, 3, 2, 3),
            LineType.LINE_2D_BOUNDARY_WEST: (0, 1, 0, 3),
            LineType.LINE_2D_BOUNDARY_EAST: (2, 1, 2, 3),
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_SOUTH: (0, 1, 2, 1),
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_NORTH: (0, 3, 2, 3),
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_WEST: (0, 1, 0, 3),
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_EAST: (2, 1, 2, 3),
        }[kcu]

    @staticmethod
    def is_horizontal(kcu: LineType) -> bool:
        return kcu in {
            LineType.LINE_2D_BOUNDARY_SOUTH,
            LineType.LINE_2D_BOUNDARY_NORTH,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_SOUTH,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_NORTH,
        }

    @staticmethod
    def is_before(kcu: LineType) -> bool:
        return kcu in {
            LineType.LINE_2D_BOUNDARY_SOUTH,
            LineType.LINE_2D_BOUNDARY_WEST,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_SOUTH,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_WEST,
        }

    @staticmethod
    def is_groundwater(kcu: LineType) -> bool:
        return kcu in {
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_SOUTH,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_NORTH,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_WEST,
            LineType.LINE_2D_GROUNDWATER_BOUNDARY_EAST,
        }

    @staticmethod
    def boundary_type_is_groundwater(boundary_type: BoundaryType) -> bool:
        return boundary_type in {
            BoundaryType.GROUNDWATERLEVEL,
            BoundaryType.GROUNDWATERDISCHARGE,
        }

    def check_edge_coord(
        self, edge_coord: np.ndarray, idx: int, is_horizontal: bool
    ) -> None:
        if edge_coord.size > 1 and np.any(edge_coord[1:] != edge_coord[:-1]):
            coords = np.unique(edge_coord)
            axis = "y" if is_horizontal else "x"
            raise SchematisationError(
                f"2D boundary condition {self.id[idx]} touches cells with "
                f"different edge coordinates ({axis}={coords.tolist()})."
            )

    def create_boundary_cells(
        self,
        idx: int,
        kcu: LineType,
        nodes: Nodes,
        node_idx: np.ndarray,
        node_id_counter: Iterator[int],
    ) -> Nodes:
        boundary_cells = nodes[node_idx]
        boundary_cells.id[:] = list(
            itertools.islice(node_id_counter, len(boundary_cells))
        )
        boundary_cells.boundary_id[:] = self.id[idx]
        boundary_cells.boundary_type[:] = self.boundary_type[idx]
        boundary_cells.node_type[:] = (
            NodeType.NODE_2D_GROUNDWATER_BOUNDARIES
            if self.is_groundwater(kcu)
            else NodeType.NODE_2D_BOUNDARIES
        )
        boundary_cells.pixel_coords[:] = -9999

        # the bounds and coordinates are shifted:
        size = boundary_cells.bounds[:, 2] - boundary_cells.bounds[:, 0]
        axes = (1, 3) if self.is_horizontal(kcu) else (0, 2)
        add_or_subtract = -1 if self.is_before(kcu) else 1
        boundary_cells.bounds[:, axes[0]] += size * add_or_subtract
        boundary_cells.bounds[:, axes[1]] += size * add_or_subtract
        boundary_cells.coordinates[:, axes[0]] += size * add_or_subtract
        if self.is_horizontal(kcu):
            boundary_cells.nodn[:] += add_or_subtract
        else:
            boundary_cells.nodm[:] += add_or_subtract
        return boundary_cells

    def create_boundary_lines(
        self,
        kcu: LineType,
        nodes: Nodes,
        node_idx: np.ndarray,
        boundary_cells: Nodes,
        line_id_counter: Iterator[int],
    ) -> Lines:
        if self.is_before(kcu):
            line = np.array([boundary_cells.id, nodes.index_to_id(node_idx)]).T
        else:
            line = np.array([nodes.index_to_id(node_idx), boundary_cells.id]).T

        cross_pix_coords = np.array(
            [
                nodes.pixel_coords[node_idx, x]
                for x in self.get_cross_pix_coords_cols(kcu)
            ]
        ).T
        return Lines(
            id=itertools.islice(line_id_counter, len(boundary_cells)),
            kcu=kcu,
            line=line,
            cross_pix_coords=cross_pix_coords,
            lik=boundary_cells.nodk,
        )

    def get_nodes_and_lines(
        self,
        nodes: Nodes,
        cell_tree: shapely.STRtree,
        quadtree,
        node_id_counter: Iterator[int],
        line_id_counter: Iterator[int],
    ) -> Tuple[Nodes, Lines]:
        # we pad quad_idx to prevent out of bound errors later
        quad_idx = np.pad(quadtree.quad_idx, 1, mode="constant", constant_values=0)

        cells_done = np.empty((0,), dtype=int)

        boundary_nodes = Nodes(id=[])
        boundary_lines = Lines(id=[])
        for bc_idx in range(len(self)):
            node_idx, is_horizontal = self.get_intersecting_node_idx(bc_idx, cell_tree)
            node_idx, is_before = self.get_neighoring_node_idx(
                bc_idx, nodes, node_idx, quad_idx, is_horizontal
            )
            is_groundwater = self.adapt_node_idx_groundwater(bc_idx, nodes, node_idx)

            kcu = self.get_kcu(is_horizontal, is_before, is_groundwater)
            edge_coord = nodes.bounds[node_idx, self.get_edge_coord_col(kcu)]
            self.check_edge_coord(edge_coord, bc_idx, is_horizontal)

            node_idx = node_idx[~np.in1d(node_idx, cells_done)]
            if len(node_idx) == 0:
                continue
            cells_done = np.append(cells_done, node_idx)

            cells_ = self.create_boundary_cells(
                bc_idx, kcu, nodes, node_idx, node_id_counter
            )
            lines_ = self.create_boundary_lines(
                kcu, nodes, node_idx, cells_, line_id_counter
            )

            boundary_nodes += cells_
            boundary_lines += lines_

        return boundary_nodes, boundary_lines



from dataclasses import dataclass
from typing import Dict, Optional

import numpy as np
import shapely

from threedigrid_builder.base import Array, surfaces
from threedigrid_builder.base.nodes import Nodes
from threedigrid_builder.constants import ContentType, SurfaceClass

__all__ = ["Surfaces", "ImperviousSurfaces"]


SURFACE_TYPE_PROPERTIES = {
    # Explanation:
    # outflow_delay: c (reaction factor) (/min)
    # surface_storage: berging (mm)
    # infiltration: boolean flag
    # fb: Max infiltration capacity (mm/h)
    # fe: Min infiltration capacity (mm/h)
    # ka: Time factor reduction (afname) of infiltration capacity (/h)
    # kh: Time factor recovery (herstel) of infiltration capacity (/h)
    # gesloten verharding, hellend
    "gvh_hel": {
        "outflow_delay": 0.5 / 60.0,
        "surface_storage": 0.0 / 1000.0,
        "infiltration": False,
        "fb": 0.0 / 60.0 / 60.0 / 1000.0,  # for consistency
        "fe": 0.0 / 60.0 / 60.0 / 1000.0,
        "ka": 0.0 / 60.0 / 60.0,
        "kh": 0.0 / 60.0 / 60.0,
    },
    # gesloten verharding, vlak
    "gvh_vla": {
        "outflow_delay": 0.2 / 60.0,
        "surface_storage": 0.5 / 1000.0,
        "infiltration": False,
        "fb": 0.0 / 60.0 / 60.0 / 1000.0,
        "fe": 0.0 / 60.0 / 60.0 / 1000.0,
        "ka": 0.0 / 60.0 / 60.0,
        "kh": 0.0 / 60.0 / 60.0,
    },
    # gesloten verharding, vlak uitgestrekt
    "gvh_vlu": {
        "outflow_delay": 0.1 / 60.0,
        "surface_storage": 1.0 / 1000.0,
        "infiltration": False,
        "fb": 0.0 / 60.0 / 60.0 / 1000.0,
        "fe": 0.0 / 60.0 / 60.0 / 1000.0,
        "ka": 0.0 / 60.0 / 60.0,
        "kh": 0.0 / 60.0 / 60.0,
    },
    # open verharding, hellend
    "ovh_hel": {
        "outflow_delay": 0.5 / 60.0,
        "surface_storage": 0.0 / 1000.0,
        "infiltration": True,
        "fb": 2.0 / 60.0 / 60.0 / 1000.0,
        "fe": 0.5 / 60.0 / 60.0 / 1000.0,
        "ka": 3.0 / 60.0 / 60.0,
        "kh": 0.1 / 60.0 / 60.0,
    },
    # open verharding, vlak
    "ovh_vla": {
        "outflow_delay": 0.2 / 60.0,
        "surface_storage": 0.5 / 1000.0,
        "infiltration": True,
        "fb": 2.0 / 60.0 / 60.0 / 1000.0,
        "fe": 0.5 / 60.0 / 60.0 / 1000.0,
        "ka": 3.0 / 60.0 / 60.0,
        "kh": 0.1 / 60.0 / 60.0,
    },
    # open verharding, vlak uitgestrekt
    "ovh_vlu": {
        "outflow_delay": 0.1 / 60.0,
        "surface_storage": 1.0 / 1000.0,
        "infiltration": True,
        "fb": 2.0 / 60.0 / 60.0 / 1000.0,
        "fe": 0.5 / 60.0 / 60.0 / 1000.0,
        "ka": 3.0 / 60.0 / 60.0,
        "kh": 0.1 / 60.0 / 60.0,
    },
    # dak, hellend
    "dak_hel": {
        "outflow_delay": 0.5 / 60.0,
        "surface_storage": 0.0 / 1000.0,
        "infiltration": False,
        "fb": 0.0 / 60.0 / 60.0 / 1000.0,
        "fe": 0.0 / 60.0 / 60.0 / 1000.0,
        "ka": 0.0 / 60.0 / 60.0,
        "kh": 0.0 / 60.0 / 60.0,
    },
    # dak, vlak
    "dak_vla": {
        "outflow_delay": 0.2 / 60.0,
        "surface_storage": 2.0 / 1000.0,
        "infiltration": False,
        "fb": 0.0 / 60.0 / 60.0 / 1000.0,
        "fe": 0.0 / 60.0 / 60.0 / 1000.0,
        "ka": 0.0 / 60.0 / 60.0,
        "kh": 0.0 / 60.0 / 60.0,
    },
    # dak, vlak uitgestrekt
    "dak_vlu": {
        "outflow_delay": 0.1 / 60.0,
        "surface_storage": 4.0 / 1000.0,
        "infiltration": False,
        "fb": 0.0 / 60.0 / 60.0 / 1000.0,
        "fe": 0.0 / 60.0 / 60.0 / 1000.0,
        "ka": 0.0 / 60.0 / 60.0,
        "kh": 0.0 / 60.0 / 60.0,
    },
    # onverhard, hellend
    "onv_hel": {
        "outflow_delay": 0.5 / 60.0,
        "surface_storage": 2.0 / 1000.0,
        "infiltration": True,
        "fb": 5.0 / 60.0 / 60.0 / 1000.0,
        "fe": 1.0 / 60.0 / 60.0 / 1000.0,
        "ka": 3.0 / 60.0 / 60.0,
        "kh": 0.1 / 60.0 / 60.0,
    },
    # onverhard, vlak
    "onv_vla": {
        "outflow_delay": 0.2 / 60.0,
        "surface_storage": 4.0 / 1000.0,
        "infiltration": True,
        "fb": 5.0 / 60.0 / 60.0 / 1000.0,
        "fe": 1.0 / 60.0 / 60.0 / 1000.0,
        "ka": 3.0 / 60.0 / 60.0,
        "kh": 0.1 / 60.0 / 60.0,
    },
    # onverhard, vlak uitgestrekt
    "onv_vlu": {
        "outflow_delay": 0.1 / 60.0,
        "surface_storage": 6.0 / 1000.0,
        "infiltration": True,
        "fb": 5.0 / 60.0 / 60.0 / 1000.0,
        "fe": 1.0 / 60.0 / 60.0 / 1000.0,
        "ka": 3.0 / 60.0 / 60.0,
        "kh": 0.1 / 60.0 / 60.0,
    },
}


SURFACE_CLASS_MAP = dict(
    [
        ("pand", "dak"),
        ("half verhard", "onv"),
        ("onverhard", "onv"),
        ("open verharding", "ovh"),
        ("gesloten verharding", "gvh"),
    ]
)

SURFACE_INCLINATION_MAP = dict(
    [("hellend", "hel"), ("vlak", "vla"), ("uitgestrekt", "vlu")]
)


@dataclass
class SurfaceParams:
    outflow_delay: np.ndarray
    surface_storage: np.ndarray
    infiltration: np.ndarray
    fb: np.ndarray
    fe: np.ndarray
    ka: np.ndarray
    kh: np.ndarray

    @classmethod
    def from_surface_class_and_inclination(
        cls, surface_class: np.ndarray, surface_inclination: np.ndarray
    ) -> "SurfaceParams":
        FIELDS = (
            ("outflow_delay", 0.0, np.float32),
            ("surface_storage", 0.0, np.float64),
            ("infiltration", False, bool),
            ("fb", 0.0, np.float64),
            ("fe", 0.0, np.float64),
            ("ka", 0.0, np.float64),
            ("kh", 0.0, np.float64),
        )

        NO_INFILTRATION_SKIP = ("fb", "fe", "ka", "kh")

        kwargs = {}
        for k, v, dtype in FIELDS:
            kwargs[k] = np.full(surface_class.shape, v, dtype)

        for sc in np.unique(surface_class):
            for si in np.unique(surface_inclination):
                sc_lookup = SURFACE_CLASS_MAP[sc]
                si_lookup = SURFACE_INCLINATION_MAP[si]
                type_props = SURFACE_TYPE_PROPERTIES[sc_lookup + "_" + si_lookup]
                mask = (surface_class == sc) & (surface_inclination == si)

                for k, _, _ in FIELDS:
                    if not type_props["infiltration"] and k in NO_INFILTRATION_SKIP:
                        continue
                    kwargs[k][mask] = type_props[k]

        return SurfaceParams(**kwargs)


class BaseSurface:
    surface_id: int
    code: str
    display_name: str
    nr_of_inhabitants: float
    area: float
    dry_weather_flow: float
    the_geom: shapely.Geometry

    connection_node_id: int
    connection_node_the_geom: shapely.Geometry
    percentage: float


class Surface(BaseSurface):
    """
    Datamodel based on "v2_surface" table
    """

    id: int
    function: str
    outflow_delay: float
    surface_layer_thickness: float
    infiltration: float
    max_infiltration_capacity: float
    min_infiltration_capacity: float
    infiltration_decay_constant: float
    infiltration_recovery_constant: float


class ImperviousSurface(BaseSurface):
    """
    Datamodel based on "v2_impervious_surface" table
    """

    id: int
    surface_inclination: str
    surface_class: SurfaceClass
    surface_sub_class: str


def fill_missing_centroids(
    centroids: np.ndarray, surface_ids: np.ndarray, connection_node_the_geom: np.ndarray
):
    """
    Fill any missing centroids with (centroids of the) connection node geometries
    """
    no_centroid_mask = shapely.is_missing(centroids)

    if np.any(no_centroid_mask):
        # Try to fill empty centroids
        empty_surface_ids = surface_ids[no_centroid_mask]
        s_unique = np.unique(empty_surface_ids)
        sort_idx = np.argsort(s_unique)
        lookup = sort_idx[np.searchsorted(s_unique, empty_surface_ids, sorter=sort_idx)]

        computed_centroids = shapely.centroid(
            shapely.multipoints(connection_node_the_geom[no_centroid_mask], lookup)
        )

        centroids[no_centroid_mask] = computed_centroids[lookup]


class BaseSurfaces:
    @property
    def unique_surfaces_mask(self) -> np.ndarray:
        if not hasattr(self, "_unique_surfaces_mask"):
            _, self._unique_surfaces_mask = np.unique(
                self.surface_id, return_index=True
            )
        return self._unique_surfaces_mask

    def as_grid_surfaces(
        self, extra_fields: Optional[Dict] = None
    ) -> surfaces.Surfaces:
        if extra_fields is None:
            extra_fields = {}

        centroids = shapely.centroid(self.the_geom)
        no_centroid_mask = shapely.is_missing(centroids)

        if np.any(no_centroid_mask):
            fill_missing_centroids(
                centroids, self.surface_id, self.connection_node_the_geom
            )

        centroid_coords = shapely.get_coordinates(centroids)

        return surfaces.Surfaces(
            id=np.arange(1, len(self.unique_surfaces_mask) + 1, dtype=int),
            code=self.code[self.unique_surfaces_mask],
            display_name=self.display_name[self.unique_surfaces_mask],
            area=self.area[self.unique_surfaces_mask],
            centroid_x=centroid_coords[:, 0][self.unique_surfaces_mask],
            centroid_y=centroid_coords[:, 1][self.unique_surfaces_mask],
            dry_weather_flow=self.dry_weather_flow[self.unique_surfaces_mask],
            nr_of_inhabitants=self.nr_of_inhabitants[self.unique_surfaces_mask],
            **extra_fields
        )

    def as_surface_maps(
        self, nodes: Nodes, nodes_embedded: Nodes
    ) -> surfaces.SurfaceMaps:
        search_array = self.surface_id[self.unique_surfaces_mask]
        sort_idx = np.argsort(self.surface_id[self.unique_surfaces_mask])
        imp_lookup = sort_idx[
            np.searchsorted(search_array, self.surface_id, sorter=sort_idx)
        ]
        connection_node_coords = shapely.get_coordinates(self.connection_node_the_geom)

        # Find connection_node (calc) node id's both in nodes and embedded nodes
        connection_node_mask = (
            nodes.content_type == ContentType.TYPE_V2_CONNECTION_NODES.value
        )
        embedded_mask = (
            nodes_embedded.content_type == ContentType.TYPE_V2_CONNECTION_NODES.value
        )

        cc_node_ids = np.concatenate(
            (
                nodes.content_pk[connection_node_mask],
                nodes_embedded.content_pk[embedded_mask],
            )
        )
        node_ids = np.concatenate(
            (nodes.id[connection_node_mask], nodes_embedded.embedded_in[embedded_mask])
        )

        sort_idx = np.argsort(cc_node_ids)
        cc_lookup = sort_idx[
            np.searchsorted(cc_node_ids, self.connection_node_id, sorter=sort_idx)
        ]

        return surfaces.SurfaceMaps(
            id=np.arange(1, len(self.connection_node_id) + 1, dtype=int),
            pk=self.connection_node_id,
            nxc=connection_node_coords[:, 0],
            nyc=connection_node_coords[:, 1],
            fac=self.percentage / 100.0,
            imp=imp_lookup,
            cci=node_ids[cc_lookup],
        )


class Surfaces(Array[Surface], BaseSurfaces):
    def as_grid_surfaces(self) -> surfaces.Surfaces:
        _, unique_surfaces_mask = np.unique(self.surface_id, return_index=True)

        infiltration_flag = self.infiltration[unique_surfaces_mask].astype("bool")
        fb = self.max_infiltration_capacity[unique_surfaces_mask] / (
            60.0 * 60.0 * 1000.0
        )
        fe = self.min_infiltration_capacity[unique_surfaces_mask] / (
            60.0 * 60.0 * 1000.0
        )
        ka = self.infiltration_decay_constant[unique_surfaces_mask] / (60.0 * 60.0)
        kh = self.infiltration_recovery_constant[unique_surfaces_mask] / (60.0 * 60.0)

        mask = np.invert(infiltration_flag)

        fb[mask] = 0.0
        fe[mask] = 0.0
        ka[mask] = 0.0
        kh[mask] = 0.0

        extra_fields = dict(
            outflow_delay=self.outflow_delay[unique_surfaces_mask] / 60.0,
            storage_limit=self.surface_layer_thickness[unique_surfaces_mask] / 1000.0,
            infiltration_flag=self.infiltration[unique_surfaces_mask].astype("bool"),
            function=self.function[unique_surfaces_mask],
            fb=fb,
            fe=fe,
            ka=ka,
            kh=kh,
        )

        return super().as_grid_surfaces(extra_fields)


class ImperviousSurfaces(Array[ImperviousSurface], BaseSurfaces):
    def as_grid_surfaces(self) -> surfaces.Surfaces:
        params = SurfaceParams.from_surface_class_and_inclination(
            self.surface_class[self.unique_surfaces_mask],
            self.surface_inclination[self.unique_surfaces_mask],
        )
        surface_sub_class = self.surface_sub_class[self.unique_surfaces_mask]
        surface_sub_class[shapely.is_missing(surface_sub_class)] = b""

        extra_fields = dict(
            surface_inclination=self.surface_inclination[self.unique_surfaces_mask],
            surface_class=self.surface_class[self.unique_surfaces_mask],
            surface_sub_class=surface_sub_class,
            function=np.full(len(self.unique_surfaces_mask), b""),
            outflow_delay=params.outflow_delay,
            storage_limit=params.surface_storage,
            infiltration_flag=params.infiltration,
            fb=params.fb,
            fe=params.fe,
            ka=params.ka,
            kh=params.kh,
        )

        return super().as_grid_surfaces(extra_fields)



from itertools import count

import pytest
from numpy.testing import assert_array_equal

from threedigrid_builder.base import Lines, Nodes
from threedigrid_builder.constants import LineType, NodeType
from threedigrid_builder.grid import Grid
from threedigrid_builder.grid import groundwater as groundwater_module


@pytest.fixture
def grid2d():
    return Grid(
        nodes=Nodes(
            id=[0, 1, 2, 3],
            node_type=NodeType.NODE_2D_OPEN_WATER,
            bounds=[(0, 0, 1, 1), (1, 0, 2, 1), (0, 1, 1, 2), (1, 1, 2, 2)],
            coordinates=[(0.5, 0.5), (1.5, 0.5), (0.5, 1.5), (1.5, 1.5)],
        ),
        lines=Lines(
            id=[0, 1, 2, 3],
            kcu=[
                LineType.LINE_2D_U,
                LineType.LINE_2D_V,
                LineType.LINE_2D_U,
                LineType.LINE_2D_OBSTACLE_V,
            ],
            line=[(0, 1), (0, 2), (1, 3), (2, 3)],
        ),
    )


@pytest.fixture
def grid2d_gw(grid2d):
    grid2d.nodes += groundwater_module.get_nodes(grid2d.nodes, count(4))[0]
    return grid2d


def test_get_nodes(grid2d):
    nodes, id_offset = groundwater_module.get_nodes(grid2d.nodes, count(6))

    assert_array_equal(nodes.id, [6, 7, 8, 9])
    assert_array_equal(nodes.node_type, NodeType.NODE_2D_GROUNDWATER)
    assert_array_equal(nodes.bounds, grid2d.nodes.bounds)
    assert id_offset == 6


def test_get_vertical_lines(grid2d_gw):
    lines = groundwater_module.get_vertical_lines(grid2d_gw.nodes, 4, count(4))

    assert_array_equal(lines.id, [4, 5, 6, 7])
    assert_array_equal(lines.kcu, LineType.LINE_2D_VERTICAL)
    assert_array_equal(lines.line, [(0, 4), (1, 5), (2, 6), (3, 7)])


def test_get_lines(grid2d_gw):
    lines = groundwater_module.get_lines(grid2d_gw.lines, 4, count(4))

    assert_array_equal(lines.id, [4, 5, 6, 7])
    assert_array_equal(lines.kcu, LineType.LINE_2D_GROUNDWATER)
    assert_array_equal(lines.line, [(4, 5), (4, 6), (5, 7), (6, 7)])



from unittest import mock

import numpy as np
import pytest
import shapely
from numpy.testing import assert_array_equal

from threedigrid_builder.base import GridSettings, Lines, Nodes, Pumps, TablesSettings
from threedigrid_builder.constants import (
    ContentType,
    InitializationType,
    LineType,
    NodeType,
)
from threedigrid_builder.grid import (
    ConnectionNodes,
    Grid,
    GridMeta,
    PotentialBreaches,
    QuadtreeStats,
)


@pytest.fixture
def connection_nodes():
    return ConnectionNodes(
        the_geom=shapely.points([(0, 0), (10, 0)]),
        id=np.array([1, 3]),
        code=np.array(["one", "two"]),
    )


@pytest.fixture
def grid():
    return Grid(nodes=Nodes(id=[]), lines=Lines(id=[]))


@pytest.fixture
def meta():
    return GridMeta(
        epsg_code=28992,
        model_name="test-name",
        grid_settings=GridSettings(
            use_2d=True,
            use_1d_flow=True,
            use_2d_flow=True,
            use_0d_inflow=0,
            grid_space=20.0,
            dist_calc_points=25.0,
            kmax=4,
        ),
        tables_settings=TablesSettings(
            table_step_size=0.05,
            frict_coef=0.03,
            frict_coef_type=9,
        ),
    )


@pytest.fixture
def grid2d(meta):
    quadtree_stats = QuadtreeStats(
        **{
            "lgrmin": 2,
            "kmax": 1,
            "mmax": np.array([1]),
            "nmax": np.array([1]),
            "dx": np.array([2.0]),
            "dxp": 0.5,
            "x0p": 10.0,
            "y0p": 10.0,
        }
    )
    return Grid(
        nodes=Nodes(
            id=[0, 1],
            node_type=NodeType.NODE_2D_OPEN_WATER,
            bounds=[(0, 0, 1, 1), (1, 0, 2, 1)],
        ),
        lines=Lines(id=[0]),
        meta=meta,
        quadtree_stats=quadtree_stats,
    )


@pytest.fixture
def grid1d(meta):
    return Grid(nodes=Nodes(id=[2, 3]), lines=Lines(id=[1]), meta=meta)


@pytest.mark.parametrize(
    "setting,expected_true",
    [
        ("interception_type", "has_interception"),
        ("groundwater_hydro_connectivity_type", "has_groundwater_flow"),
        ("infiltration_rate_type", "has_simple_infiltration"),
        ("groundwater_impervious_layer_level_type", "has_groundwater"),
        ("interflow_type", "has_interflow"),
    ],
)
def test_from_meta(meta, setting, expected_true):
    setattr(meta.tables_settings, setting, InitializationType.GLOBAL)
    grid = Grid.from_meta(
        epsg_code=28992,
        model_name="test",
        grid_settings=meta.grid_settings,
        tables_settings=meta.tables_settings,
    )
    for attr in (
        "has_interception",
        "has_groundwater_flow",
        "has_simple_infiltration",
        "has_groundwater",
        "has_interflow",
    ):
        assert getattr(grid.meta, attr) is (attr == expected_true)


def test_set_crs_pyproj(grid2d, crs_wkt_28992):
    grid2d.meta.epsg_code = None
    grid2d.meta.crs_wkt = None

    grid2d.set_crs(crs_wkt_28992)
    assert grid2d.meta.epsg_code == 28992
    assert grid2d.meta.crs_wkt == crs_wkt_28992.srs


def test_set_crs_gdal(grid2d, crs_wkt_28992_legacy):
    grid2d.meta.epsg_code = None
    grid2d.meta.crs_wkt = None
    grid2d.set_crs(crs_wkt_28992_legacy)
    assert grid2d.meta.epsg_code == 28992
    assert grid2d.meta.crs_wkt == crs_wkt_28992_legacy.srs


def test_from_quadtree():
    quadtree = mock.Mock()
    area_mask = mock.Mock()
    counter = mock.Mock()
    nodes = Nodes(id=[])
    lines = Lines(id=[])

    quadtree.origin = (0.0, 0.0)
    quadtree.get_nodes_lines.return_value = nodes, lines

    grid = Grid.from_quadtree(
        quadtree=quadtree,
        area_mask=area_mask,
        node_id_counter=counter,
        line_id_counter=counter,
    )

    assert isinstance(grid, Grid)
    assert grid.nodes is nodes
    assert grid.lines is lines


def test_from_connection_nodes():
    connection_nodes = mock.Mock()
    counter = mock.Mock()
    nodes = Nodes(id=[])

    connection_nodes.get_nodes.return_value = nodes
    grid = Grid.from_connection_nodes(connection_nodes, counter)

    connection_nodes.get_nodes.assert_called_with(counter)

    assert grid.nodes is nodes
    assert len(grid.lines) == 0


def test_concatenate_grid(grid2d, grid1d):
    grid = grid2d + grid1d
    assert grid.meta == grid1d.meta
    assert grid.quadtree_stats == grid2d.quadtree_stats
    assert_array_equal(grid.nodes.id[0:2], grid2d.nodes.id)
    assert_array_equal(grid.nodes.id[2:], grid1d.nodes.id)


@mock.patch("threedigrid_builder.grid.connection_nodes.set_calculation_types")
def test_set_calculation_types(set_calculation_types, grid):
    grid.set_calculation_types()
    set_calculation_types.assert_called_with(grid.nodes, grid.lines)


@mock.patch("threedigrid_builder.grid.connection_nodes.set_bottom_levels")
def test_set_bottom_levels(cn_compute):
    CH = ContentType.TYPE_V2_CHANNEL
    CN = ContentType.TYPE_V2_CONNECTION_NODES
    PI = ContentType.TYPE_V2_PIPE
    CV = ContentType.TYPE_V2_CULVERT

    # set an input grid and mock the compute_weights return value
    nodes = Nodes(
        id=[1, 2, 3, 4, 5, 6, 7],
        content_type=[CH, -9999, CH, PI, CV, CN, CN],
        content_pk=[1, 1, 1, 2, 5, 1, 2],
        s1d=[2.0, 12.0, 21.0, 15.0, 0.5, np.nan, np.nan],
        dmax=[np.nan, 12.0, np.nan, np.nan, np.nan, np.nan, np.nan],
    )
    lines = Lines(
        id=range(7),
        content_type=[CH, CH, CH, PI, PI, CV, CV],
        line=[(6, 1), (1, 3), (3, 7), (6, 4), (4, 7), (6, 5), (5, 7)],
        invert_level_start_point=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
        invert_level_end_point=[2.0, 3.0, 3.0, 5.0, 6.0, 7.0, 8.0],
    )
    grid = Grid(nodes=nodes, lines=lines)
    grid.set_bottom_levels()

    # node attribute dmax is adapted correctly
    assert_array_equal(grid.nodes.dmax, [2.0, 12.0, 3.0, 5.0, 7.0, np.nan, np.nan])

    # connection node set_bottom_levels was called correctly
    cn_compute.assert_called_with(grid.nodes, grid.lines)


@mock.patch("threedigrid_builder.grid.initial_waterlevels.compute_initial_waterlevels")
def test_set_initial_waterlevels(compute_initial_waterlevels, grid):
    connection_nodes = mock.Mock()
    channels = mock.Mock()
    pipes = mock.Mock()
    culverts = mock.Mock()

    grid.set_initial_waterlevels(connection_nodes, channels, pipes, culverts)

    compute_initial_waterlevels.assert_called_with(
        grid.nodes,
        connection_nodes=connection_nodes,
        channels=channels,
        pipes=pipes,
        culverts=culverts,
    )


@pytest.fixture
def grid_for_sorting():
    return Grid(
        Nodes(
            id=[0, 1, 2, 3, 4, 5],
            node_type=[
                NodeType.NODE_1D_BOUNDARIES,
                NodeType.NODE_2D_OPEN_WATER,
                NodeType.NODE_2D_OPEN_WATER,
                NodeType.NODE_2D_OPEN_WATER,
                NodeType.NODE_1D_STORAGE,
                NodeType.NODE_1D_NO_STORAGE,
            ],
            dmax=[0, 1, 2, 3, 4, 5],
        ),
        Lines(
            id=[0, 1, 6],
            kcu=[
                LineType.LINE_1D_SHORT_CRESTED,
                LineType.LINE_1D_ISOLATED,
                LineType.LINE_1D2D_SINGLE_CONNECTED_OPEN_WATER,
            ],
            is_1d_boundary=[1, -9999, -9999],
            line=[(0, 4), (4, 5), (5, 1)],
            dpumax=[0, 1, 6],
        ),
        pumps=Pumps(id=[0], line=[(4, 5)]),
        nodes_embedded=Nodes(id=[0], embedded_in=[1]),
        breaches=PotentialBreaches(id=[2], line_id=[6]),
    )


def test_sort(grid_for_sorting):
    grid = grid_for_sorting
    grid.sort()

    assert_array_equal(grid.nodes.id, [0, 1, 2, 3, 4, 5])
    assert_array_equal(grid.nodes.dmax, [1, 2, 3, 4, 5, 0])
    assert_array_equal(grid.lines.id, [0, 1, 2])
    assert_array_equal(grid.lines.dpumax, [1, 6, 0])
    assert_array_equal(grid.lines.line, [(3, 4), (4, 0), (5, 3)])
    assert_array_equal(grid.pumps.line, [(3, 4)])
    assert_array_equal(grid.nodes_embedded.embedded_in, [0])
    assert_array_equal(grid.breaches.line_id, [1])


def test_sort_no_lines(grid_for_sorting):
    grid_for_sorting.lines = grid_for_sorting.lines[:0]
    grid_for_sorting.pumps = grid_for_sorting.pumps[:0]
    grid_for_sorting.breaches = grid_for_sorting.breaches[:0]

    grid_for_sorting.sort()


def test_sort_null_lines_err(grid_for_sorting):
    grid_for_sorting.lines.line[0] = (-9999, 2)
    with pytest.raises(ValueError):
        grid_for_sorting.sort()


def test_sort_boundary_conditions():
    grid = Grid(
        Nodes(
            id=[0, 1, 2, 3, 4, 5, 6, 7, 8],
            node_type=[
                NodeType.NODE_2D_OPEN_WATER,
                NodeType.NODE_1D_BOUNDARIES,
                NodeType.NODE_2D_BOUNDARIES,
                NodeType.NODE_2D_BOUNDARIES,
                NodeType.NODE_1D_BOUNDARIES,
                NodeType.NODE_2D_BOUNDARIES,
                NodeType.NODE_1D_NO_STORAGE,
                NodeType.NODE_2D_GROUNDWATER_BOUNDARIES,
                NodeType.NODE_2D_GROUNDWATER_BOUNDARIES,
            ],
            dmax=[0, 1, 2, 3, 4, 5, 6, 7, 8],
        ),
        Lines(
            id=[0, 1, 2, 3, 4, 5, 6],
            kcu=[
                LineType.LINE_1D_ISOLATED,
                LineType.LINE_1D_ISOLATED,
                LineType.LINE_2D_BOUNDARY_EAST,
                LineType.LINE_2D_BOUNDARY_WEST,
                LineType.LINE_2D_BOUNDARY_NORTH,
                LineType.LINE_1D2D_SINGLE_CONNECTED_CLOSED,
                LineType.LINE_2D_GROUNDWATER_BOUNDARY_NORTH,
            ],
            is_1d_boundary=[1, 1] + [-9999] * 5,
            line=[(6, 4), (1, 6), (0, 3), (0, 5), (2, 0), (0, 6), (7, 8)],
            dpumax=[0, 1, 2, 3, 4, 5, 6],
        ),
    )
    grid.sort()

    # expected order:
    #   2D (0), 1D (6)
    #   2D boundaries (2, 3, 5)
    #   2D groundwater (7, 8)
    #   1D boundaries (1, 4)
    assert_array_equal(grid.nodes.id, [0, 1, 2, 3, 4, 5, 6, 7, 8])
    assert_array_equal(grid.nodes.dmax, [0, 6, 2, 3, 5, 7, 8, 1, 4])
    # expected order:
    #   1D2D (5)
    #   2D boundaries (4, 2, 3)
    #   2D Groundwater (6)
    #   1D boundaries (1, 0)
    # note that the 1D and 2D boundaries are sorted internally so that the order matches
    # the order in the corresponding nodes
    assert_array_equal(grid.lines.id, [0, 1, 2, 3, 4, 5, 6])
    assert_array_equal(grid.lines.dpumax, [5, 4, 2, 3, 6, 1, 0])
    # the correct line order can also be seen by the increasing node ids:
    assert_array_equal(
        grid.lines.line, [[0, 1], [2, 0], [0, 3], [0, 4], [5, 6], [7, 1], [1, 8]]
    )


def test_set_cross_sections(grid):
    definitions = mock.Mock()
    grid.lines = Lines(
        id=[0, 1, 2, 3], cross_id1=[2, 5, 2, -9999], cross_id2=[2, -9999, 3, 4]
    )

    grid.set_cross_sections(definitions)

    assert_array_equal(definitions.convert.call_args[0][0], [2, 3, 4, 5])
    assert grid.cross_sections is definitions.convert.return_value



import numpy as np
import shapely

from threedigrid_builder.base import Array, Lines


class Obstacle:
    id: int
    crest_level: float
    the_geom: shapely.Geometry


class Obstacles(Array[Obstacle]):
    def compute_dpumax(self, lines: Lines, where):
        """Compute the dpumax for lines that intersect with the obstacles.

        Args:
            lines (Lines)
            where (int array): indices into lines to compute dpumax for

        Returns:
            dpumax for each line (length equals length of 'where' array)
        """
        obstacle_idx = np.full(len(where), -9999, dtype=np.int32)
        exchange_level = np.full(len(where), np.nan)
        if len(self) == 0:
            return exchange_level, obstacle_idx
        line_geometries = lines.line_geometries[where]
        if shapely.is_missing(line_geometries).any():
            raise ValueError(
                "Obstacles.compute_dpumax requires line_geometries to be set"
            )
        lines_tree = shapely.STRtree(lines.line_geometries[where])
        inscts = lines_tree.query(self.the_geom, predicate="intersects")
        for i in range(len(self)):
            line_idx = inscts[1, np.where(inscts[0, :] == i)[0]]
            is_greater = ~(self.crest_level[i] <= exchange_level[line_idx])
            exchange_level[line_idx[is_greater]] = self.crest_level[i]
            obstacle_idx[line_idx[is_greater]] = i
        return exchange_level, obstacle_idx



import logging

import numpy as np
import shapely

from threedigrid_builder.base import (
    Array,
    LineHalfs,
    Lines,
    Nodes,
    PointsOnLine,
    search,
)
from threedigrid_builder.constants import CalculationType, ContentType, Material
from threedigrid_builder.exceptions import SchematisationError

from .channels import Channels

__all__ = [
    "PotentialBreaches",
    "PotentialBreachPoints",
]

logger = logging.getLogger(__name__)


class PotentialBreach:
    id: int
    the_geom: shapely.Geometry
    code: str
    display_name: str
    exchange_level: float
    levee_material: Material
    maximum_breach_depth: float
    channel_id: int
    line_id: int  # only used for the output
    content_pk: int  # only used for the output


class PotentialBreachPoints(PointsOnLine):
    def merge(self) -> "PotentialBreachPoints":
        """Merge breach points that are on the same position.

        - Breaches are merged into channel start / end if they are exactly on it.
        - Breaches are merged with one another if they are exactly on the same place.
        """
        s1d = self.s1d[:]

        eps = 1e-4

        # snap breaches to channel starts/ends (to fix numerical imprecisions)
        s1d[s1d < eps] = 0.0
        lengths = self.linestrings.length[self.linestring_idx]
        mask = lengths - s1d < eps
        s1d[mask] = lengths[mask]

        # compute interdistances and a list of points that are the same
        dists = np.diff(s1d)
        same_channel = np.diff(self.linestring_idx) == 0
        same = np.where((dists < eps) & same_channel)[0]
        to_delete = same + 1  # deleting these leaves only 'primary points'

        # for each primary point keep track of the secondary one
        secondary_content_pk = np.full(len(self), fill_value=-9999, dtype=np.int32)
        for rng in np.split(same, np.where(np.diff(same) != 1)[0] + 1):
            if len(rng) == 0:
                continue
            secondary_content_pk[rng[0]] = self.content_pk[rng[0] + 1]

        # get the second one of a too close pair
        return self.__class__(
            linestrings=self.linestrings,
            id=np.delete(self.id, to_delete),
            content_pk=np.delete(self.content_pk, to_delete),
            secondary_content_pk=np.delete(secondary_content_pk, to_delete),
            linestring_idx=np.delete(self.linestring_idx, to_delete),
            s1d=np.delete(s1d, to_delete),
        )

    def find_for_line_halfs(self, line_halfs: LineHalfs):
        """Return a breach for each line_half by matching against channel id.

        It is assumed there is exactly 1 breach per line_half. This is ensured
        by PotentialBreachPoints.merge().
        """
        # per (is_start) line_half, look for a breach point (at start)
        idx = np.full(len(line_halfs), fill_value=-9999, dtype=np.int32)
        idx[line_halfs.is_start] = search(
            self.linestring_id,  # channel id
            line_halfs.content_pk[line_halfs.is_start],  # lines.content_pk
            mask=self.at_start,
            check_exists=False,
        )
        # per (is_end) line_half, look for a breach point (at end)
        idx[line_halfs.is_end] = search(
            self.linestring_id,
            line_halfs.content_pk[line_halfs.is_end],
            mask=self.at_end,
            check_exists=False,
        )
        has_breach_point = idx != -9999
        idx = idx[has_breach_point]
        return line_halfs.node_id[has_breach_point], idx

    def assign_to_connection_nodes(self, nodes: Nodes, lines: Lines):
        """Per connection node, assign max two potential breach ids.

        The priority is as follows:
        - Take the breach points of the first channel that has 2 breach
          points at the node.
        - If there are no double breach points: take the breach points of
          the first channel.
        """
        # disassemble lines into Channel - Connection Node line_halfs
        line_halfs = LineHalfs.for_connection_nodes(
            nodes, lines, line_mask=lines.content_type == ContentType.TYPE_V2_CHANNEL
        )
        line_halfs.reorder_by("node_id")

        # per line_half, match a breach point by their channel ids
        node_ids, breach_point_idx = self.find_for_line_halfs(line_halfs)
        node_idx = nodes.id_to_index(node_ids)

        # iterate over nodes with a breach point and assign (if present)
        # the first double breach point
        for node_idx_ in np.unique(node_idx):
            options = breach_point_idx[node_idx == node_idx_]
            # sort by linestring (channel) id for reproducibility:
            options = options[np.argsort(self.linestring_id[options])]
            breach_id_1 = self.content_pk[options]
            breach_id_2 = self.secondary_content_pk[options]
            is_double = np.where(breach_id_2 != -9999)[0]
            idx = is_double[0] if len(is_double) > 0 else 0
            nodes.breach_ids[node_idx_] = [breach_id_1[idx], breach_id_2[idx]]

    @staticmethod
    def match_breach_ids_with_calculation_types(nodes: Nodes):
        """Check if the number of breach ids on a point matches its calculation type.

        - max 1 breach for CONNECTED
        - max 2 breaches for DOUBLE_CONNECTED
        - no breaches otherwise
        """
        has_a_breach = nodes.breach_ids[:, 0] != -9999
        has_2_breach = nodes.breach_ids[:, 1] != -9999
        is_double_connected = nodes.calculation_type == CalculationType.DOUBLE_CONNECTED
        is_single_connected = nodes.calculation_type == CalculationType.CONNECTED

        one_is_too_much = np.where(
            has_a_breach & ~(is_single_connected | is_double_connected)
        )[0]
        if len(one_is_too_much) > 0:
            raise SchematisationError(
                f"The following objects have potential breaches, but are not "
                f"(double) connected: {nodes.format_message(one_is_too_much)}."
            )
        two_is_too_much = np.where(has_2_breach & ~is_double_connected)[0]
        if len(two_is_too_much) > 0:
            raise SchematisationError(
                f"The following objects have two potential breaches at the "
                f"same position, but are not double connected: "
                f"{nodes.format_message(two_is_too_much)}."
            )


class PotentialBreaches(Array[PotentialBreach]):
    @property
    def side_1d(self):
        return shapely.get_point(self.the_geom, 0)

    @property
    def side_2d(self):
        return shapely.get_point(self.the_geom, -1)

    def project_on_channels(self, channels: Channels) -> PotentialBreachPoints:
        """Project the potential breaches on channels, yielding points on channels.

        This method also calls the 'merge' logic.
        """
        return PotentialBreachPoints.from_geometries(
            channels.linestrings,
            points=self.side_1d,
            linestring_idx=channels.id_to_index(self.channel_id, check_exists=True),
            content_pk=self.id,
        )



import numpy as np
import shapely
from osgeo import gdal, gdal_array, ogr

from threedigrid_builder.base import Array


class GridRefinement:
    id: int
    display_name: str
    code: str
    refinement_level: int
    the_geom: shapely.Geometry


class GridRefinements(Array[GridRefinement]):
    def rasterize(self, **kwargs):
        """Return a raster with refinement_level at cells where a refinement intersect.

        When multiple refinements intersect the same cell, the lowest refinement level
        is taken.
        """
        # sort by refinement_level (lowest values last, so that they are on top)
        sorter = np.argsort(self.refinement_level)[::-1]
        return rasterize(self.the_geom[sorter], self.refinement_level[sorter], **kwargs)


def rasterize(geoms, values, origin, width, height, cell_size, no_data_value):
    """Rasterize a set of geometries to a raster with values at cells where the
    geometries intersect.

    Args:
        arr (array of shapely.Geometry)
        values (array of int)
        origin (tuple of 2 floats): x, y order
        width (int)
        height (int)
        cell_size (float)
        no_data_value (int)

    Returns:
        array with shape (y, x) and dtype int32
    """
    assert np.issubdtype(values.dtype, np.integer)
    assert height > 0
    assert width > 0
    dtype = np.int32  # OGR knows int32 and int64, but GDAL only int32
    ogr_dtype = ogr.OFTInteger

    # initialize the array
    array = np.full((1, height, width), no_data_value, dtype=dtype)

    # drop empty geometries
    mask = shapely.is_geometry(geoms)
    geoms = geoms[mask]
    values = values[mask]

    # if there are no features, return directly
    if len(geoms) == 0:
        return array[0]

    # create an output datasource in memory
    driver = ogr.GetDriverByName("Memory")
    burn_attr = "BURN_IT"

    # prepare in-memory ogr layer
    ds_ogr = driver.CreateDataSource("")
    layer = ds_ogr.CreateLayer("")
    layer_definition = layer.GetLayerDefn()
    field_definition = ogr.FieldDefn(burn_attr, ogr_dtype)
    layer.CreateField(field_definition)

    for geometry, value in zip(geoms, values):
        feature = ogr.Feature(layer_definition)
        feature.SetGeometry(ogr.CreateGeometryFromWkb(shapely.to_wkb(geometry)))
        if ogr_dtype is not None:
            feature[burn_attr] = value
        layer.CreateFeature(feature)

    dataset_kwargs = {
        "no_data_value": no_data_value,
        "geo_transform": [
            float(x) for x in (origin[0], cell_size, 0, origin[1], 0, cell_size)
        ],
    }

    # ATTRIBUTE=BURN_ATTR burns the BURN_ATTR value of each feature
    # ALL_TOUCHED turns off the special GDAL algorithm for lines
    options = ["ATTRIBUTE=" + burn_attr, "ALL_TOUCHED=1"]

    with Dataset(array, **dataset_kwargs) as dataset:
        gdal.RasterizeLayer(dataset, (1,), layer, options=options)

    return array[0]


def create_dataset(array, geo_transform=None, projection=None, no_data_value=None):
    """
    Create and return a gdal dataset.

    :param array: A numpy array.
    :param geo_transform: 6-tuple of floats
    :param projection: wkt projection string
    :param no_data_value: integer or float

    This is the fastest way to get a gdal dataset from a numpy array, but
    keep a reference to the array around, or a segfault will occur. Also,
    don't forget to call FlushCache() on the dataset after any operation
    that affects the array.
    """
    # prepare dataset name pointing to array
    datapointer = array.ctypes.data
    bands, lines, pixels = array.shape
    datatypecode = gdal_array.NumericTypeCodeToGDALTypeCode(array.dtype.type)
    datatype = gdal.GetDataTypeName(datatypecode)
    bandoffset, lineoffset, pixeloffset = array.strides

    dataset_name_template = (
        "MEM:::"
        "DATAPOINTER={datapointer},"
        "PIXELS={pixels},"
        "LINES={lines},"
        "BANDS={bands},"
        "DATATYPE={datatype},"
        "PIXELOFFSET={pixeloffset},"
        "LINEOFFSET={lineoffset},"
        "BANDOFFSET={bandoffset}"
    )
    dataset_name = dataset_name_template.format(
        datapointer=datapointer,
        pixels=pixels,
        lines=lines,
        bands=bands,
        datatype=datatype,
        pixeloffset=pixeloffset,
        lineoffset=lineoffset,
        bandoffset=bandoffset,
    )

    # access the array memory as gdal dataset
    dataset = gdal.Open(dataset_name, gdal.GA_Update)

    # set additional properties from kwargs
    if geo_transform is not None:
        dataset.SetGeoTransform(geo_transform)
    if projection is not None:
        dataset.SetProjection(projection)
    if no_data_value is not None:
        for i in range(len(array)):
            dataset.GetRasterBand(i + 1).SetNoDataValue(no_data_value)

    return dataset


class Dataset(object):
    """
    Usage:
        >>> with Dataset(array) as dataset:
        ...     # do gdal things.
    """

    def __init__(self, array, **kwargs):
        self.array = array
        self.dataset = create_dataset(array, **kwargs)

    def __enter__(self):
        return self.dataset

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.dataset.FlushCache()



from itertools import count

import pytest
import shapely
from numpy.testing import assert_almost_equal, assert_array_equal

from threedigrid_builder.base import Array, Lines, Nodes
from threedigrid_builder.constants import CalculationType, ContentType, NodeType
from threedigrid_builder.exceptions import SchematisationError
from threedigrid_builder.grid import (
    ConnectionNodes,
    embed_linear_objects,
    embed_nodes,
    Grid,
)
from threedigrid_builder.grid.linear import BaseLinear

NODE_2D = NodeType.NODE_2D_OPEN_WATER
NODE_1D = NodeType.NODE_1D_STORAGE
EMBEDDED = CalculationType.EMBEDDED
ISOLATED = CalculationType.ISOLATED


class LinearObject:
    id: int
    the_geom: shapely.Geometry
    calculation_type: CalculationType
    connection_node_start_id: int
    connection_node_end_id: int
    display_name: str


class LinearObjects(Array[LinearObject], BaseLinear):
    content_type = ContentType.TYPE_V2_WINDSHIELD  # just pick one for the test


@pytest.fixture
def grid2d():
    return Grid(
        nodes=Nodes(
            id=[0, 1, 2, 3],
            node_type=NODE_2D,
            bounds=[(0, 0, 10, 10), (10, 0, 20, 10), (0, 10, 10, 20), (10, 10, 20, 20)],
        ),
        lines=Lines(id=[0], line=[[0, 1]]),
    )


def test_embed_nodes(grid2d):
    grid = grid2d + Grid(
        nodes=Nodes(
            id=[4, 5, 6, 7],
            node_type=NODE_1D,
            content_type=ContentType.TYPE_V2_CONNECTION_NODES,
            calculation_type=[EMBEDDED, ISOLATED, EMBEDDED, EMBEDDED],
            coordinates=[(10, 2), (9, 9), (13, 3), (16, 7)],
        ),
        lines=Lines(id=[1, 2, 3], line=[(4, 5), (4, 6), (5, 6)]),
    )
    embedded = embed_nodes(grid, count(10))

    assert_array_equal(grid.nodes.id, [0, 1, 2, 3, 4])
    assert_array_equal(grid.nodes.node_type, [NODE_2D] * 4 + [NODE_1D])
    assert_array_equal(grid.lines.line, [(0, 1), (0, 4), (0, 1), (4, 1)])

    assert_array_equal(embedded.id, [10, 11, 12])  # new ids
    assert_array_equal(embedded.calculation_type, EMBEDDED)
    assert_array_equal(embedded.embedded_in, [0, 1, 1])


def test_embed_node_outside_2D(grid2d):
    grid = grid2d + Grid(
        nodes=Nodes(
            id=[4, 5],
            node_type=NODE_1D,
            content_type=ContentType.TYPE_V2_CONNECTION_NODES,
            content_pk=[15, 16],
            calculation_type=EMBEDDED,
            coordinates=[(5, 22), (5, 5)],
        ),
        lines=Lines(id=[]),
    )

    with pytest.raises(SchematisationError, match=r".*\[15\] are outside the 2D cells"):
        embed_nodes(grid, count(10))


def test_embed_node_two_interconnected(grid2d):
    grid = grid2d + Grid(
        nodes=Nodes(
            id=[4, 5],
            node_type=NODE_1D,
            content_type=ContentType.TYPE_V2_CONNECTION_NODES,
            content_pk=[4, 5],
            calculation_type=EMBEDDED,
            coordinates=[(8, 8), (5, 5)],
        ),
        lines=Lines(id=[2], line=[(4, 5)]),
    )

    with pytest.raises(SchematisationError, match=r".*\[4, 5\] connect to.*"):
        embed_nodes(grid, count(10))


@pytest.fixture
def connection_nodes():
    # Used to map connection_node_start/end_id to an index (sequence id)
    return ConnectionNodes(
        id=[21, 25, 33, 42],
        the_geom=shapely.points([(0, 21), (1, 25), (2, 33), (3, 42)]),
    )


def test_embed_linear_objects_multiple(grid2d):
    linear_objects = LinearObjects(
        id=[0, 1, 2, 3],
        calculation_type=EMBEDDED,
        the_geom=[
            shapely.linestrings([(1, 1), (15, 1)]),
            shapely.linestrings([(11, 1), (15, 1)]),
            shapely.linestrings([(6, 7), (10, 10), (15, 10), (15, 15), (10, 15)]),
            shapely.linestrings([(7, 2), (18, 2), (18, 9), (18, 18)]),
        ],
    )

    nodes, lines_s1d, lines_ds1d_half = embed_linear_objects(
        linear_objects,
        grid2d.cell_tree,
        embedded_cutoff_threshold=1,
        embedded_node_id_counter=count(2),
    )

    assert_array_equal(nodes.id, [2])
    assert_array_equal(nodes.content_type, ContentType.TYPE_V2_WINDSHIELD)
    assert_array_equal(nodes.content_pk, [3])
    assert_array_equal(nodes.coordinates, [(18, 2)])
    assert_array_equal(nodes.calculation_type, EMBEDDED)
    assert_array_equal(nodes.s1d, [11])  # halfway the 2 velocity points (see below)
    assert_array_equal(nodes.embedded_in, [1])
    assert_almost_equal(lines_s1d, [9, 2, 5, 3, 19])
    assert_almost_equal(lines_ds1d_half, [9, 2, 5, 3, 8])


@pytest.mark.parametrize("reverse", [False, True])
@pytest.mark.parametrize(
    "geometry,lines_s1d,embedded_in",
    [
        ([(5, 5), (9, 5)], [2], []),  # no cell crossing --> no nodes and 1 line
        ([(5, 5), (18, 5)], [5], []),  # horizontal, 1 crossing
        ([(5, 5), (5, 18)], [5], []),  # vertical, 1 crossing
        ([(5, 5), (5, 9), (15, 9)], [9], []),  # 1 crossing, more coords
        ([(5, 5), (18, 5), (18, 9)], [5], []),  # 1 crossing, more coords
        ([(5, 15), (5, 1), (18, 1)], [5, 19], [0]),  # corner, bottomleft
        ([(5, 5), (18, 5), (18, 13)], [5, 18], [1]),  # corner, bottomright
        ([(5, 5), (5, 14), (18, 14)], [5, 14], [2]),  # corner, bottomright
        ([(18, 5), (18, 13), (9, 13)], [5, 16], [3]),  # corner, topright
        ([(5, 15), (5, 2), (19, 2), (19, 15)], [5, 18, 35], [0, 1]),  # U
        ([(5, 2), (19, 2), (19, 15), (5, 15)], [5, 22, 36], [1, 3]),  # U, on left side
        ([(19, 2), (19, 15), (5, 15), (5, 2)], [8, 22, 32], [3, 2]),  # U, upside down
        ([(19, 15), (5, 15), (5, 2), (19, 2)], [9, 19, 32], [2, 0]),  # U, on right side
        ([(8, 1), (12, 4), (8, 7)], [2.5, 7.5], [1]),  # 0 - 1 - 0
        ([(8, 1), (12, 4), (8, 7), (12, 7)], [2.5, 7.5, 12.0], [1, 0]),  # 0 - 1 - 0 - 1
        ([(5, 5), (18, 5), (18, 10)], [5], []),  # 1 crossing, end at edge
        ([(5, 15), (5, 1), (18, 1), (18, 10)], [5, 19], [0]),  # corner, end at edge
        ([(6, 1), (10, 4), (6, 7)], [5.0], []),  # 0 - touch 1
        ([(14, 1), (10, 4), (14, 7)], [5.0], []),  # 1 - touch 0
        ([(12, 7), (6, 7), (10, 4), (6, 1)], [2], []),  # 2 - 0 - touch 1
        ([(6, 1), (10, 4), (6, 7), (12, 7)], [14], []),  # 0 - touch 1 - 2
        ([(1, 5), (1, 14), (4, 10), (7, 14), (7, 5)], [5, 23], [2]),  # M (0 - 2)
        ([(5, 2), (10, 2), (10, 8), (5, 8)], None, []),  # 0 - along edge - 0
        ([(5, 2), (10, 2), (10, 8), (15, 8)], None, []),  # 0 - along edge - 1
        ([(5, 2), (10, 2), (10, 14), (14, 14)], None, []),  # 0 - along edge - 2
        ([(5, 2), (10, 2), (10, 10), (8, 10), (8, 8)], None, []),
        ([(10, 2), (10, 5), (5, 5)], None, []),  # start at edge
        ([(10, 2), (10, 5), (5, 5), (5, 15)], None, []),  # start at edge - 0 - 2
        ([(10, 2), (10, 5), (5, 5), (5, 15), (15, 15)], None, [2]),
        ([(5, 5), (5, 10), (7, 10), (7, 15), (10, 15), (10, 17), (15, 17)], None, [2]),
        ([(5.6, 7), (10, 10.3), (15, 10.3)], [5.25], []),  # 0 to 3, 2 is below thresh
        ([(5.6, 7), (10, 10.3), (14.4, 7)], [5.5], []),  # 0 to 1, 2&3 are below thresh
        ([(0, 5), (18, 5)], [10], []),  # begins at model edge
        ([(0, 5), (0, 15), (5, 15)], [5], []),  # begins tangent to the model edge
        ([(0, 5), (18, 5), (18, 12)], [10, 23], [1]),  # begins at model edge, 2
        ([(0, 5), (0, 15), (15, 15)], [5, 20], [2]),  # begins tangent to model edge, 2
    ],
)
def test_embed_linear_object(grid2d, geometry, lines_s1d, embedded_in, reverse):
    if reverse:
        geometry = geometry[::-1]
        embedded_in = embedded_in[::-1]
        if lines_s1d is not None:
            lines_s1d = shapely.length(shapely.linestrings(geometry)) - lines_s1d[::-1]

    linear_objects = LinearObjects(
        id=[0],
        calculation_type=EMBEDDED,
        the_geom=[shapely.linestrings(geometry)],
    )

    nodes, actual_lines_s1d, line_ds1d_half = embed_linear_objects(
        linear_objects,
        grid2d.cell_tree,
        embedded_cutoff_threshold=1,
        embedded_node_id_counter=count(2),
    )

    assert_array_equal(nodes.embedded_in, embedded_in)
    if lines_s1d is not None:
        assert_almost_equal(actual_lines_s1d, lines_s1d)


@pytest.mark.parametrize("reverse", [False, True])
@pytest.mark.parametrize(
    "geometry",
    [
        [(-1, 5), (18, 5)],  # begins outside of model
        [(5, 5), (-5, 5), (-5, 15), (5, 15)],  # begins in the model, but goes outside
    ],
)
def test_embed_linear_object_outside_raise(grid2d, geometry, reverse):
    if reverse:
        geometry = geometry[::-1]

    linear_objects = LinearObjects(
        id=[0],
        calculation_type=EMBEDDED,
        the_geom=[shapely.linestrings(geometry)],
    )

    with pytest.raises(
        SchematisationError,
        match=r"LinearObjects \[0\] are not completely inside the 2D cell.",
    ):
        embed_linear_objects(
            linear_objects,
            grid2d.cell_tree,
            embedded_cutoff_threshold=1,
            embedded_node_id_counter=count(2),
        )



import numpy as np
import pytest
from numpy.testing import assert_equal

from threedigrid_builder.base import Lines
from threedigrid_builder.constants import ContentType
from threedigrid_builder.grid import Windshieldings


@pytest.fixture
def channel_lines():
    # ordered by [content_pk, position on channel]
    return Lines(
        id=[0, 1, 2, 3, 4, 5, 6],
        content_type=ContentType.TYPE_V2_CHANNEL,
        content_pk=[51, 52, 53, 53, 53, 54, 54],
        line=[(0, 1), (0, 2), (0, 3), (3, 4), (4, 5), (0, 6), (6, 7)],
    )


@pytest.fixture
def mixed_lines():
    # ordered by [content_pk, position on channel]
    return Lines(
        id=[0, 1, 2, 3, 4, 5, 6],
        content_type=[
            ContentType.TYPE_V2_PIPE,
            ContentType.TYPE_V2_CULVERT,
            ContentType.TYPE_V2_CHANNEL,
            ContentType.TYPE_V2_CHANNEL,
            ContentType.TYPE_V2_CHANNEL,
            ContentType.TYPE_V2_WEIR,
            ContentType.TYPE_V2_WEIR,
        ],
        content_pk=[53, 53, 53, 53, 53, 53, 53],
        line=[(0, 1), (0, 2), (0, 3), (3, 4), (4, 5), (0, 6), (6, 7)],
    )


@pytest.fixture
def different_lines():
    # ordered by [content_pk, position on channel]
    return Lines(
        id=[0, 1, 2, 3, 4, 5, 6],
        content_type=[
            ContentType.TYPE_V2_PIPE,
            ContentType.TYPE_V2_CULVERT,
            ContentType.TYPE_V2_CHANNEL,
            ContentType.TYPE_V2_CHANNEL,
            ContentType.TYPE_V2_CHANNEL,
            ContentType.TYPE_V2_WEIR,
            ContentType.TYPE_V2_WEIR,
        ],
        content_pk=[53, 53, 1, 2, 3, 53, 53],
        line=[(0, 1), (0, 2), (0, 3), (3, 4), (4, 5), (0, 6), (6, 7)],
    )


@pytest.fixture
def windshieldings():
    north = np.array([0.1, 0.2, 0.5, 0.4], dtype=np.float64)
    east = np.array([0.0, 0.1, 0.3, 0.9], dtype=np.float64)
    south = np.array([0.3, 0.4, 0.1, 0.8], dtype=np.float64)
    west = np.array([0.4, 0.3, 0.2, 0.1], dtype=np.float64)
    return Windshieldings(
        id=[1, 2, 3, 4],
        channel_id=[51, 53, 52, 54],
        north=north,
        northeast=north + 1,
        east=east,
        southeast=east + 1,
        south=south,
        southwest=south + 1,
        west=west,
        northwest=west + 1,
    )


@pytest.fixture
def windshieldings_channel_selection():
    north = np.array([0.1], dtype=np.float64)
    east = np.array([0.0], dtype=np.float64)
    south = np.array([0.3], dtype=np.float64)
    west = np.array([0.4], dtype=np.float64)
    return Windshieldings(
        id=[1],
        channel_id=[51],
        north=north,
        northeast=north + 1,
        east=east,
        southeast=east + 1,
        south=south,
        southwest=south + 1,
        west=west,
        northwest=west + 1,
    )


def test_set_windshielding(channel_lines, windshieldings):
    windshieldings.apply_to_lines(channel_lines)
    expected_windshieldings = np.array(
        [
            [0.1, 1.1, 0.0, 1.0, 0.3, 1.3, 0.4, 1.4],
            [0.5, 1.5, 0.3, 1.3, 0.1, 1.1, 0.2, 1.2],
            [0.2, 1.2, 0.1, 1.1, 0.4, 1.4, 0.3, 1.3],
            [0.2, 1.2, 0.1, 1.1, 0.4, 1.4, 0.3, 1.3],
            [0.2, 1.2, 0.1, 1.1, 0.4, 1.4, 0.3, 1.3],
            [0.4, 1.4, 0.9, 1.9, 0.8, 1.8, 0.1, 1.1],
            [0.4, 1.4, 0.9, 1.9, 0.8, 1.8, 0.1, 1.1],
        ],
        dtype=np.float64,
    )
    assert_equal(channel_lines.windshieldings, expected_windshieldings)


def test_set_windshielding_no_windshieldings(channel_lines, windshieldings):
    Windshieldings(id=[]).apply_to_lines(channel_lines)
    expected_windshieldings = np.full((7, 8), np.nan, dtype=np.float64)
    assert_equal(channel_lines.windshieldings, expected_windshieldings)


def test_set_windshielding_different_lines(mixed_lines, windshieldings):
    windshieldings.apply_to_lines(mixed_lines)
    expected_windshieldings = np.array(
        [
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            [0.2, 1.2, 0.1, 1.1, 0.4, 1.4, 0.3, 1.3],
            [0.2, 1.2, 0.1, 1.1, 0.4, 1.4, 0.3, 1.3],
            [0.2, 1.2, 0.1, 1.1, 0.4, 1.4, 0.3, 1.3],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        ],
        dtype=np.float64,
    )
    assert_equal(mixed_lines.windshieldings, expected_windshieldings)


def test_set_windshielding_channel_selection(
    channel_lines, windshieldings_channel_selection
):
    windshieldings_channel_selection.apply_to_lines(channel_lines)
    expected_windshieldings = np.array(
        [
            [0.1, 1.1, 0.0, 1.0, 0.3, 1.3, 0.4, 1.4],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        ],
        dtype=np.float64,
    )
    assert_equal(channel_lines.windshieldings, expected_windshieldings)



import itertools
from typing import Optional

import numpy as np
import shapely

from threedigrid_builder.base import Lines, LineStrings, Nodes, PointsOnLine
from threedigrid_builder.base.linestrings import counts_to_ranges
from threedigrid_builder.constants import CalculationType, NodeType

COORD_EQUAL_ATOL = 1e-8  # the distance below which coordinates are considered equal


class BaseLinear:
    content_type = None  # to be defined by subclasses

    @property
    def linestrings(self):
        return LineStrings(objects=self)

    def set_geometries(self, connection_nodes):
        """Set the_geom from connection nodes where necessary.

        This also reverses geometries if necessary.

        Args:
            connection_nodes (ConnectionNodes): to take the coordinates from
        """
        # get the start and end points
        points_1 = connection_nodes.the_geom[
            connection_nodes.id_to_index(self.connection_node_start_id)
        ]
        points_2 = connection_nodes.the_geom[
            connection_nodes.id_to_index(self.connection_node_end_id)
        ]

        self.linestrings.sanitize(points_1, points_2)

        # construct the geometries where necessary
        has_no_geom = shapely.is_missing(self.the_geom)
        coordinates = np.empty((np.count_nonzero(has_no_geom), 2, 2))
        coordinates[:, 0, 0] = shapely.get_x(points_1[has_no_geom])
        coordinates[:, 0, 1] = shapely.get_y(points_1[has_no_geom])
        coordinates[:, 1, 0] = shapely.get_x(points_2[has_no_geom])
        coordinates[:, 1, 1] = shapely.get_y(points_2[has_no_geom])
        self.the_geom[has_no_geom] = shapely.linestrings(coordinates)

    def interpolate_nodes(
        self,
        node_id_counter,
        global_dist_calc_points,
        fixed_nodes: Optional[PointsOnLine] = None,
    ):
        """Compute nodes on each linear object with constant intervals

        The following fields are expected to be filled on self:
        - id
        - the_geom
        - dist_calc_points
        - calculation_type

        Args:
            node_id_counter (iterable): an iterable yielding integers
            global_dist_calc_points (float): Default node interdistance.
            fixed_nodes: optional fixed points that will become a node

        Returns:
            tuple of nodes (Nodes)
            nodes has data in the following columns:
            - id: counter generated from node_id_counter
            - coordinates: (x, y) coordinates of the node
            - content_pk: the id of the linear object from which the node originates
            - node_type: NodeType.NODE_1D_NO_STORAGE
            - calculation_type: the calculation type copied from the linear object
            - s1d: distance (along the linestring) to the start of the linestring
            The nodes are ordered by content_pk and then by position on the linestring.
        """
        if shapely.is_missing(self.the_geom).any():
            raise ValueError(
                f"{self.__class__.__name__} encountered without a geometry."
            )

        # normalize global dist_calc_points
        if (
            global_dist_calc_points is None
            or np.isnan(global_dist_calc_points)
            or global_dist_calc_points <= 0.0
        ):
            global_dist_calc_points = np.inf  # means: no interpolation

        if fixed_nodes is None:
            fixed_nodes = PointsOnLine.empty(self.linestrings)
        else:
            fixed_nodes = fixed_nodes[~(fixed_nodes.at_start | fixed_nodes.at_end)]

        # insert default dist_calc_points where necessary
        dists = self.dist_calc_points.copy()
        dists[np.isnan(dists)] = global_dist_calc_points
        dists[dists <= 0] = global_dist_calc_points
        # Do not add nodes for embedded objects:
        dists[self.calculation_type == CalculationType.EMBEDDED] = np.inf

        # interpolate the node geometries
        sublinestrings = self.linestrings.segmentize(fixed_nodes)
        points = sublinestrings.interpolate_points(dists[sublinestrings.linestring_idx])

        # fixed_nodes also become nodes
        points = points.merge_with(fixed_nodes)

        # construct the nodes with available attributes
        return Nodes(
            id=itertools.islice(node_id_counter, len(points)),
            coordinates=shapely.get_coordinates(points.the_geom),
            content_type=self.content_type,
            content_pk=self.id[points.linestring_idx],
            node_type=NodeType.NODE_1D_NO_STORAGE,
            calculation_type=self.calculation_type[points.linestring_idx],
            s1d=points.s1d,
            breach_ids=np.array([points.content_pk, points.secondary_content_pk]).T,
        )

    def get_embedded(
        self, cell_tree, embedded_cutoff_threshold, embedded_node_id_counter
    ):
        """ """
        from threedigrid_builder.grid import embed_linear_objects

        return embed_linear_objects(
            self[self.calculation_type == CalculationType.EMBEDDED],
            cell_tree,
            embedded_cutoff_threshold,
            embedded_node_id_counter,
        )

    def get_lines(
        self,
        connection_nodes,
        nodes,
        line_id_counter,
        connection_node_offset=0,
        embedded_mode=False,
    ):
        """Compute the grid lines for the linear objects.

        The following fields are expected to be filled on self:
        - id
        - the_geom
        - connection_node_start_id
        - connection_node_end_id
        - calculation_type
        - cross_section_definition_id (pipes and culverts only)
        - display_name

        Args:
            connection_nodes (ConnectionNodes): used to map ids to indices
            nodes (Nodes): interpolated nodes (see interpolate_nodes)
            line_id_counter (iterable): an iterable yielding integers
            connection_node_offset (int): offset to give connection node
              indices in the returned lines.line. Default 0.
            embedded_mode (bool): whether to get lines for embedded objects
              if embedded_mode is true, the line.line come from node.embedded_in

        Returns:
            Lines with data in the following columns:
            - id: counter generated from line_id_counter
            - line: 2 node ids per line (or a different attribute, see line_id_attr)
            - content_pk: the id of the linear from which this line originates
            - s1d: the positon of the line center along the linear object
            - ds1d: the arclength of the line
            - kcu: the calculation_type of the linear object
            - line_geometries: the linestrings (segments of self.the_geom)
            These attributes are filled for pipes/culverts only:
            - cross_id1 & cross_id2: the id of the cross section definition
            - cross_weight: 1.0 (which means that cross_id2 should be ignored)
            - frict_type1 & frict_type2: the friction type (both are equal)
            - frict_value1 & frict_value2: the friction value (both are equal)
            - invert_level_start_point: copied from pipe/culvert
            - invert_level_end_point: copied from pipe/culvert
            - dpumax: largest of the two invert levels
            - discharge_coefficient_positive & _positive: culverts only
            The lines are ordered by content_pk and then by position on the linestring.
        """
        if embedded_mode:
            objs = self[self.calculation_type == CalculationType.EMBEDDED]
        else:
            objs = self[self.calculation_type != CalculationType.EMBEDDED]

        # Cut the linestrings up into segments (LinesOnLine)
        node_points = PointsOnLine.from_s1d(
            objs.linestrings,
            nodes.s1d,
            objs.id_to_index(nodes.content_pk),
            content_pk=nodes.content_pk,
        )
        segments = objs.linestrings.segmentize(node_points)
        segment_idx = segments.linestring_idx

        # Copy properties to segments
        display_name = np.take(objs.display_name, segment_idx)
        zoom_category = np.take(objs.zoom_category, segment_idx)
        connection_node_start_id = np.take(objs.connection_node_start_id, segment_idx)
        connection_node_end_id = np.take(objs.connection_node_end_id, segment_idx)
        dist_calc_points = np.take(objs.dist_calc_points, segment_idx)

        # set the right node indices for each segment
        first_idx, last_idx = counts_to_ranges(np.bincount(segments.linestring_idx))
        last_idx -= 1  # convert slice end into last index
        line = np.full((len(segments), 2), -9999, dtype=np.int32)

        # convert connection_node_start_id to index and put it at first segments' start
        line[first_idx, 0] = (
            connection_nodes.id_to_index(objs.connection_node_start_id)
            + connection_node_offset
        )
        # convert connection_node_end_id to index and put it at last segments' end
        line[last_idx, 1] = (
            connection_nodes.id_to_index(objs.connection_node_end_id)
            + connection_node_offset
        )
        # set node indices to line start where segment start is not a conn. node
        mask = np.ones(len(segments), dtype=bool)
        mask[first_idx] = False
        line[mask, 0] = nodes.id if not embedded_mode else nodes.embedded_in
        # set node indices to line end where segment end is not a conn. node
        mask = np.ones(len(segments), dtype=bool)
        mask[last_idx] = False
        line[mask, 1] = nodes.id if not embedded_mode else nodes.embedded_in

        # map to obj ids
        content_pk = objs.id[segment_idx]

        # conditionally add the cross section definition (for pipes and culverts only)
        try:
            cross_id = objs.cross_section_definition_id[segment_idx]
            cross_weight = 1.0
        except AttributeError:
            cross_id = -9999
            cross_weight = np.nan

        # conditionally add the invert levels (for pipes and culverts only)
        try:
            invert_start = self.compute_bottom_level(content_pk, segments.s1d_start)
            invert_end = self.compute_bottom_level(content_pk, segments.s1d_end)
            dpumax = np.maximum(invert_start, invert_end)
        except AttributeError:
            invert_end = invert_start = dpumax = np.nan

        # conditionally add friction type and value (for pipes and culverts only)
        try:
            frict_type = objs.friction_type[segment_idx]
            frict_value = objs.friction_value[segment_idx]
        except AttributeError:
            frict_type = -9999
            frict_value = np.nan

        # conditionall add hydraulic conductivity (for channels and pipes only)
        try:
            exchange_thickness = objs.exchange_thickness
            hydraulic_resistance_in = (
                objs.hydraulic_conductivity_in[segment_idx]
                / exchange_thickness[segment_idx]
            )
            hydraulic_resistance_out = (
                objs.hydraulic_conductivity_out[segment_idx]
                / exchange_thickness[segment_idx]
            )
        except AttributeError:
            hydraulic_resistance_in = np.nan
            hydraulic_resistance_out = np.nan

        # Conditionally add discharge coefficients, (for culverts only). If one culvert has
        # multiple segments positive coefficient goes onto the first segment and negative
        # coefficient goes onto last segment (otherwise we have to much energy losses.)
        try:
            dc_positive = np.full((len(segments)), 1.0, dtype=np.float64)
            dc_negative = np.full((len(segments)), 1.0, dtype=np.float64)
            dc_positive[first_idx] = objs.discharge_coefficient_positive
            dc_negative[last_idx] = objs.discharge_coefficient_negative
        except AttributeError:
            dc_positive = 1.0
            dc_negative = 1.0

        # Conditionally add material (pipes only)
        try:
            material = objs.material[segment_idx]
        except AttributeError:
            material = -9999

        # Conditionally add sewerage_type (pipes only)
        try:
            sewerage_type = objs.sewerage_type[segment_idx]
        except AttributeError:
            sewerage_type = -9999

        # construct the result
        return Lines(
            id=itertools.islice(line_id_counter, len(segments)),
            line_geometries=segments.the_geom,
            line=line,
            content_type=objs.content_type,
            content_pk=content_pk,
            s1d=segments.s1d,
            ds1d_half=segments.ds1d / 2,
            ds1d=segments.ds1d,
            kcu=objs.calculation_type[segment_idx],
            cross_id1=cross_id,
            cross_id2=cross_id,
            cross_weight=cross_weight,
            frict_type1=frict_type,
            frict_value1=frict_value,
            frict_type2=frict_type,
            frict_value2=frict_value,
            invert_level_start_point=invert_start,
            invert_level_end_point=invert_end,
            dpumax=dpumax,
            discharge_coefficient_positive=dc_positive,
            discharge_coefficient_negative=dc_negative,
            display_name=display_name,
            zoom_category=zoom_category,
            connection_node_start_id=connection_node_start_id,
            connection_node_end_id=connection_node_end_id,
            dist_calc_points=dist_calc_points,
            material=material,
            sewerage_type=sewerage_type,
            hydraulic_resistance_in=hydraulic_resistance_in,
            hydraulic_resistance_out=hydraulic_resistance_out,
        )

    def compute_bottom_level(self, ids, s):
        """Compute the bottom level by interpolating between invert levels

        This function is to be used on nodes and lines between on pipes and culverts.

        Args:
            ids (ndarray of int): the id (content_pk) of the object
            s (ndarray of float): the position of the node/line along the object

        Returns:
            an array of the same shape as ids and ds containing the interpolated values
        """
        if shapely.is_missing(self.the_geom).any():
            raise ValueError(
                f"{self.__class__.__name__} found without a geometry. Call "
                f"set_geometries first."
            )
        lengths = shapely.length(self.the_geom)
        idx = self.id_to_index(ids)
        weights = s / lengths[idx]
        if np.any(weights < 0.0) or np.any(weights > 1.0):
            raise ValueError("Encountered nodes outside of the linear object bounds")

        left = self.invert_level_start_point[idx]
        right = self.invert_level_end_point[idx]

        return weights * right + (1 - weights) * left

    def compute_drain_level(self, ids, s, connection_nodes):
        """Compute the drain level by interpolating between Manhole drain levels

        This function is to be used for interpolated nodes on pipes and culverts.

        Args:
            ids (ndarray of int): the id (content_pk) of the object
            s (ndarray of float): the position of the node measured along the object
            connection_nodes (ConnectionNodes): to obtain drain levels

        Returns:
            an array of the same shape as ids and ds containing the interpolated values
        """
        if shapely.is_missing(self.the_geom).any():
            raise ValueError(
                f"{self.__class__.__name__} found without a geometry. Call "
                f"set_geometries first."
            )
        lengths = shapely.length(self.the_geom)
        idx = self.id_to_index(ids)
        weights = s / lengths[idx]
        if np.any(weights < 0.0) or np.any(weights > 1.0):
            raise ValueError("Encountered nodes outside of the linear object bounds")

        # convert connection_node_start_id to index to get the drain level
        left_cn_idx = connection_nodes.id_to_index(self.connection_node_start_id)[idx]
        right_cn_idx = connection_nodes.id_to_index(self.connection_node_end_id)[idx]

        left = connection_nodes.drain_level[left_cn_idx]
        right = connection_nodes.drain_level[right_cn_idx]

        result = weights * right + (1 - weights) * left

        # fix cases with nan on one side
        result[np.isnan(result)] = left[np.isnan(result)]
        result[np.isnan(result)] = right[np.isnan(result)]
        return result

    @property
    def has_groundwater_exchange(self):
        return np.full(len(self), False, dtype=bool)

    def apply_has_groundwater_exchange(self, nodes: Nodes, lines: Lines):
        content_pk = self.id[self.has_groundwater_exchange]

        if len(content_pk) == 0:
            return

        node_ids = np.unique(
            lines.line[
                (lines.content_type == self.content_type)
                & np.isin(lines.content_pk, content_pk)
            ]
        )

        nodes.has_groundwater_exchange[
            np.isin(nodes.id, node_ids)
            & (nodes.calculation_type != CalculationType.BOUNDARY_NODE)
        ] = True



from unittest import mock

import numpy as np
import pytest
import shapely
from shapely.testing import assert_geometries_equal

from threedigrid_builder.base import GridSettings, Pumps, TablesSettings
from threedigrid_builder.constants import (
    BoundaryType,
    CalculationType,
    CrossSectionShape,
    FrictionType,
    InitializationType,
)
from threedigrid_builder.exceptions import SchematisationError
from threedigrid_builder.grid import (
    BoundaryConditions1D,
    BoundaryConditions2D,
    Channels,
    ConnectionNodes,
    CrossSectionDefinitions,
    CrossSectionLocations,
    Culverts,
    Obstacles,
    Orifices,
    Pipes,
    Weirs,
)
from threedigrid_builder.interface import SQLite


def test_init(tmp_path):
    path = tmp_path / "some.sqlite"
    path.touch()

    with mock.patch(
        "threedigrid_builder.interface.db.ThreediDatabase"
    ) as db, mock.patch.object(SQLite, "get_version") as get_version:
        get_version.return_value = 216
        sqlite = SQLite(path)

    db.assert_called_with(path)

    assert sqlite.db is db.return_value


def test_init_no_file(tmp_path):
    path = tmp_path / "some.sqlite"

    with pytest.raises(FileNotFoundError):
        SQLite(path)


def test_init_bad_version(tmp_path):
    path = tmp_path / "some.sqlite"
    path.touch()

    with mock.patch(
        "threedigrid_builder.interface.db.ThreediDatabase"
    ), mock.patch.object(SQLite, "get_version") as get_version:
        with pytest.raises(SchematisationError):
            get_version.return_value = 171
            SQLite(path)


def test_get_version(db):
    assert db.get_version() == 216


def test_get_boundary_conditions_1d(db):
    boundary_conditions_1d = db.get_boundary_conditions_1d()
    assert isinstance(boundary_conditions_1d, BoundaryConditions1D)

    # some test samples
    assert len(boundary_conditions_1d) == 4
    assert boundary_conditions_1d.id[1] == 2
    assert boundary_conditions_1d.boundary_type[2] == BoundaryType.DISCHARGE
    assert boundary_conditions_1d.connection_node_id[3] == 59


def test_get_boundary_conditions_2d(db):
    boundary_conditions_2d = db.get_boundary_conditions_2d()
    assert isinstance(boundary_conditions_2d, BoundaryConditions2D)

    assert len(boundary_conditions_2d) == 0


def test_get_channels(db):
    channels = db.get_channels()
    assert isinstance(channels, Channels)

    # some test samples
    assert len(channels.id) == 1175
    assert_geometries_equal(
        channels.the_geom[0],
        shapely.from_wkt(
            "LINESTRING (109798 518896, 109815 518823, 109818 518812, 109822 518800, 109834 518750)"
        ),
        tolerance=1,
    )
    assert channels.id[11] == 12
    assert channels.code[42] == "151"
    assert channels.calculation_type[494] == CalculationType.CONNECTED
    assert channels.dist_calc_points[580] == 100.0
    assert channels.connection_node_start_id[536] == 1377
    assert channels.connection_node_end_id[1163] == 1056
    assert channels.id[1174] == 666668899
    assert channels.display_name[33] == "801"


def test_get_connection_nodes(db):
    connection_nodes = db.get_connection_nodes()
    assert isinstance(connection_nodes, ConnectionNodes)

    # some test samples
    assert len(connection_nodes.id) == 1360
    assert_geometries_equal(
        connection_nodes.the_geom[0],
        shapely.from_wkt("POINT (110404 517792)"),
        tolerance=1,
    )
    assert connection_nodes.id[11] == 12
    assert connection_nodes.storage_area[39] == 0.64
    assert np.isnan(connection_nodes.storage_area[49])
    assert connection_nodes.code[494] == ""
    assert connection_nodes.initial_waterlevel[169] == -0.4
    # manhole fields
    assert connection_nodes.manhole_id[10] == 11
    assert connection_nodes.manhole_id[100] == -9999
    assert connection_nodes.calculation_type[1] == CalculationType.CONNECTED
    assert connection_nodes.manhole_indicator[6] == 1
    assert connection_nodes.bottom_level[9] == -3.51
    assert connection_nodes.drain_level[1] == -0.82
    assert connection_nodes.surface_level[35] == -0.54
    assert connection_nodes.shape[40] == "00"
    assert connection_nodes.width[32] == 0.8
    assert connection_nodes.display_name[33] == "71512"


def test_get_cross_section_definitions(db):
    definitions = db.get_cross_section_definitions()
    assert isinstance(definitions, CrossSectionDefinitions)

    # some test samples
    assert len(definitions.id) == 11
    assert definitions.id[8] == 97
    assert definitions.shape[7] == CrossSectionShape.RECTANGLE
    assert definitions.height[10] == "0.4"
    assert definitions.width[2] == "0.315"


def test_get_cross_section_locations(db):
    locations = db.get_cross_section_locations()
    assert isinstance(locations, CrossSectionLocations)

    # some test samples
    assert len(locations.id) == 1175
    assert_geometries_equal(
        locations.the_geom[96], shapely.from_wkt("POINT (111104 521655)"), tolerance=1
    )
    assert locations.id[11] == 12
    assert locations.definition_id[365] == 98
    assert locations.channel_id[448] == 452
    assert locations.reference_level[691] == -3.0
    assert locations.bank_level[995] == -1.7
    assert locations.friction_type[1103] == FrictionType.MANNING
    assert locations.friction_value[1103] == 0.03


def test_get_grid_refinements(db):
    grid_refinements = db.get_grid_refinements()

    # some test samples
    assert len(grid_refinements.id) == 6
    assert_geometries_equal(
        grid_refinements.the_geom[3],
        shapely.from_wkt(
            "LINESTRING (110173 517604, 110327 517527, 110461 517809, 110249 517909, 110147 517700, 110304 517616, 110368 517765, 110280 517816, 110242 517726, 110300 517701, 110316 517749)"
        ),
        tolerance=1,
    )
    assert_geometries_equal(
        grid_refinements.the_geom[4],
        shapely.from_wkt(
            "POLYGON ((108334 517481, 108701 517460, 108686 517171, 108324 517197, 108334 517481))"
        ),
        tolerance=1,
    )
    assert grid_refinements.refinement_level[1] == 1
    assert grid_refinements.display_name[3] == "riolering"
    assert grid_refinements.display_name[5] == "test_polygon2"
    assert grid_refinements.code[5] == "2"


def test_get_obstacles(db):
    obstacles = db.get_obstacles()
    assert isinstance(obstacles, Obstacles)

    assert len(obstacles) == 3
    assert obstacles.id[1] == 2
    assert_geometries_equal(
        shapely.get_point(obstacles.the_geom[0], 0),
        shapely.from_wkt("POINT (110241 519070)"),
        tolerance=1,
    )
    assert obstacles.crest_level[1] == 0.0


def test_get_pipes(db):
    pipes = db.get_pipes()
    assert isinstance(pipes, Pipes)

    # some test samples
    assert len(pipes.id) == 42
    assert pipes.id[1] == 2
    assert pipes.code[2] == "71022_71023"
    assert pipes.calculation_type[3] == CalculationType.ISOLATED
    assert pipes.connection_node_start_id[4] == 37
    assert pipes.connection_node_end_id[5] == 35
    assert pipes.cross_section_definition_id[9] == 7
    assert pipes.invert_level_start_point[16] == -3.91
    assert pipes.invert_level_end_point[19] == -3.62
    assert pipes.sewerage_type[24] == 2
    assert pipes.friction_type[28] == FrictionType.MANNING
    assert pipes.friction_value[36] == 0.0145
    assert pipes.display_name[33] == "71518_71517"
    assert pipes.zoom_category[15] == 3


def test_get_settings(db):
    result = db.get_settings()
    assert result["epsg_code"] == 28992
    assert result["model_name"] == "simple_infil_no_grndwtr"

    g = result["grid_settings"]
    assert isinstance(g, GridSettings)
    # some settings copied over from SQLite:
    assert g.use_1d_flow is True
    assert g.use_2d_flow is True
    assert g.grid_space == 20.0
    assert g.dist_calc_points == 15.0
    assert g.kmax == 4
    assert g.embedded_cutoff_threshold == 0.05
    assert g.max_angle_1d_advection == 0.4 * np.pi
    # use_2d is based on the presence of dem_file:
    assert g.use_2d is True

    s = result["tables_settings"]
    assert isinstance(s, TablesSettings)
    assert s.table_step_size == 0.05
    assert s.frict_type == 2
    assert s.frict_coef == 0.03
    assert s.frict_coef_type == InitializationType.GLOBAL
    assert s.interception_global == 100.0
    assert s.interception_type == InitializationType.NO_AGG
    assert s.table_step_size_1d == 0.05
    assert s.maximum_table_step_size == 5.0
    assert s.manhole_storage_area is None

    # groundwater settings
    assert s.groundwater_impervious_layer_level == -5.0
    assert s.groundwater_impervious_layer_level_type == InitializationType.GLOBAL
    assert s.phreatic_storage_capacity == 0.25
    assert s.phreatic_storage_capacity_type == InitializationType.GLOBAL
    assert s.equilibrium_infiltration_rate == 100.0
    assert s.equilibrium_infiltration_rate_type == InitializationType.GLOBAL
    assert s.initial_infiltration_rate == 300.0
    assert s.initial_infiltration_rate_type == InitializationType.GLOBAL
    assert s.infiltration_decay_period == 0.1
    assert s.infiltration_decay_period_type == InitializationType.GLOBAL
    assert s.groundwater_hydro_connectivity == 1.0
    assert s.groundwater_hydro_connectivity_type == InitializationType.GLOBAL

    # there are interflow settings, but most are unset
    assert s.interflow_type == 0  # means: no interflow
    assert s.porosity is None
    assert s.porosity_type is None
    assert s.porosity_layer_thickness is None
    assert s.impervious_layer_elevation is None
    assert s.hydraulic_conductivity is None
    assert s.hydraulic_conductivity_type is None

    # there are no infiltration settings
    assert s.infiltration_rate is None
    assert s.infiltration_rate_type is None
    assert s.infiltration_surface_option == 0
    assert s.max_infiltration_capacity is None
    assert s.max_infiltration_capacity_type is None

    # there are no vegetation_drag settings
    assert s.vegetation_height is None
    assert s.vegetation_height_type is None
    assert s.vegetation_stem_count is None
    assert s.vegetation_stem_count_type is None
    assert s.vegetation_stem_diameter is None
    assert s.vegetation_stem_diameter_type is None
    assert s.vegetation_drag_coefficient is None
    assert s.vegetation_drag_coefficient is None


def test_get_pumps(db):
    pumps = db.get_pumps()
    assert isinstance(pumps, Pumps)

    # some test samples
    assert len(pumps) == 19
    assert pumps.id[11] == 13
    assert pumps.code[0] == "Rioolgemaal"
    assert pumps.capacity[12] == 0.288
    assert pumps.connection_node_start_id[13] == 1006
    assert pumps.connection_node_end_id[0] == -9999  # NULL handling
    assert pumps.connection_node_end_id[2] == 218
    assert pumps.type_[5] == 1
    assert pumps.start_level[0] == -4.0
    assert pumps.lower_stop_level[18] == -1.9
    assert np.isnan(pumps.upper_stop_level[15])
    assert pumps.zoom_category[15] == 5
    assert pumps.display_name[10] == "KGM-JL-18"


def test_get_culverts(db):
    culverts = db.get_culverts()
    assert isinstance(culverts, Culverts)

    # some test samples
    assert len(culverts) == 92

    assert culverts.id[89] == 2000112
    assert culverts.code[35] == "500"
    assert_geometries_equal(
        culverts.the_geom[36],
        shapely.from_wkt("LINESTRING (108351 516428, 108357 516430)"),
        tolerance=1,
    )
    assert culverts.dist_calc_points[52] == 20.0
    assert culverts.connection_node_start_id[45] == 1220
    assert culverts.connection_node_end_id[61] == 1620
    assert culverts.calculation_type[4] == CalculationType.ISOLATED
    assert culverts.cross_section_definition_id[0] == 97
    assert culverts.invert_level_start_point[21] == -2.39
    assert culverts.invert_level_end_point[83] == -3.28
    assert culverts.discharge_coefficient_negative[0] == 0.8
    assert culverts.discharge_coefficient_positive[0] == 0.8
    assert culverts.friction_type[28] == FrictionType.MANNING
    assert culverts.friction_value[36] == 0.03
    assert culverts.display_name[32] == "KDU-Q-18482"
    assert culverts.zoom_category[15] == 2


def test_get_orifices(db):
    orifices = db.get_orifices()
    assert isinstance(orifices, Orifices)

    # no orifices in test dataset
    assert len(orifices) == 0


def test_get_weirs(db):
    weirs = db.get_weirs()
    assert isinstance(weirs, Weirs)

    # some test samples
    assert len(weirs) == 56

    assert weirs.id[13] == 17
    assert weirs.code[11] == "317"
    assert weirs.connection_node_start_id[40] == 1643
    assert weirs.connection_node_end_id[7] == 394
    assert weirs.crest_level[26] == -2.508
    assert weirs.crest_type[0] == CalculationType.SHORT_CRESTED
    assert weirs.cross_section_definition_id[0] == 8
    assert weirs.discharge_coefficient_negative[0] == 0.0
    assert weirs.discharge_coefficient_positive[0] == 0.8
    assert weirs.friction_type[28] == FrictionType.MANNING
    assert weirs.friction_value[36] == 0.03
    assert weirs.display_name[33] == "KST-JL-76"
    assert weirs.zoom_category[15] == 2
    assert weirs.sewerage[0] == 1


def test_get_dem_average(db):
    dem_avg_areas = db.get_dem_average_areas()
    # no dem_average_areas in test dataset
    assert len(dem_avg_areas) == 0


def test_get_surfaces(db):
    surfaces = db.get_surfaces()
    # No surfaces in test dataset
    assert len(surfaces) == 0


def test_get_impervious_surfaces(db):
    impervious_surfaces = db.get_impervious_surfaces()
    # No impervious_surfaces in test dataset
    assert len(impervious_surfaces) == 0


def test_get_windshieldings(db):
    windshieldings = db.get_windshieldings()
    # No windshielding in test dataset
    assert len(windshieldings) == 0


def test_get_potential_breaches(db):
    potential_breaches = db.get_potential_breaches()
    assert len(potential_breaches) == 2


def test_get_exchange_lines(db):
    exchange_lines = db.get_exchange_lines()
    # No exchange lines in test dataset
    assert len(exchange_lines) == 0



import os
import pathlib
import shutil
import sys

from setuptools import find_packages

try:
    from skbuild import setup
except ImportError:
    from setuptools import setup
    if not any(x in sys.argv for x in {"sdist", "--version", "egg_info"}):
        sys.exit("ERROR: skbuild is required to build threedigrid-builder from source.")

ext_modules = []

if "clean" in sys.argv:
    # delete any previously compiled files
    sys.argv.remove("clean")
    p = pathlib.Path(".")
    for pattern in [
        "libthreedigrid/*.c",
        "libthreedigrid/*.f",
        "libthreedigrid/*.mod",
        "threedigrid_builder/grid/fgrid/*.so",
    ]:
        for filename in p.glob(pattern):
            print("removing '{}'".format(filename))
            filename.unlink()
    print("removing build folder")
    if os.path.isdir(p / "_skbuild"):
        shutil.rmtree(p / "_skbuild")

if sys.platform == "win32":
    cmake_args = ["-G", "MSYS Makefiles", "-DCMAKE_GNUtoMS=ON"]
else:
    cmake_args = []

long_description = open("README.rst").read()


def get_version():
    # Edited from https://packaging.python.org/guides/single-sourcing-package-version/
    init_path = pathlib.Path(__file__).parent / "threedigrid_builder/__init__.py"
    for line in init_path.open("r").readlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


install_requires = [
    "numpy>=1.15",
    "threedi-schema>=0.216.0",
    "shapely>=2",
    "pyproj>=3",
    "condenser[geo]>=0.1.1",
    "sqlalchemy>=1.4.1",
]

test_requires = ["pytest", "pytest-cov"]

setup(
    name="threedigrid-builder",
    version=get_version(),
    description="Generate a 3Di simulation grid from a model schematisation.",
    long_description=long_description,
    url="https://docs.3di.lizard.net/",
    author="Martijn Siemerink",
    author_email="martijn.siemerink@nelen-schuurmans.nl",
    license="GNU General Public License v3.0",
    packages=find_packages(
        include=(
            "threedigrid_builder",
            "threedigrid_builder.*",
        ),
    ),
    cmake_args=cmake_args,
    install_requires=install_requires,
    extras_require={
        "test": test_requires,
        "gridadmin": ["h5py>=2.7"],
        "gpkg": ["geopandas"],
        "cli": ["typer"],
    },
    python_requires=">=3.7",
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "License :: Other/Proprietary License",
    ],
    zip_safe=False,
    entry_points={
        "console_scripts": ["threedigrid-builder=threedigrid_builder.cli:run"],
    },
)



import os
import pathlib
import shutil

import numpy as np
import pytest
import shapely
from pyproj import CRS

from threedigrid_builder.base import (
    GridSettings,
    Lines,
    Nodes,
    Pumps,
    Surfaces,
    TablesSettings,
)
from threedigrid_builder.base.surfaces import SurfaceMaps
from threedigrid_builder.constants import ContentType, NodeType
from threedigrid_builder.grid import (
    CrossSections,
    Grid,
    GridMeta,
    Obstacles,
    PotentialBreaches,
    QuadtreeStats,
)
from threedigrid_builder.interface.db import SQLite

data_path = pathlib.Path(__file__).resolve().parents[0] / "data"


@pytest.fixture(scope="session")
def db(tmp_path_factory):
    """Yields a threedigrid_builder.interface.db.SQLite object with access
    to the test v2_bergermeer.sqlite."""
    fn = tmp_path_factory.mktemp("data") / "v2_bergermeer.sqlite"
    sqlite_path = data_path / "v2_bergermeer.sqlite"
    shutil.copyfile(sqlite_path, fn)
    if not os.path.isfile(fn):
        pytest.skip("sample sqlite is not available", allow_module_level=True)
    return SQLite(fn, upgrade=True)


@pytest.fixture
def dem_path():
    """Yields a dem path"""
    return data_path / "dem_test_5m.tif"


@pytest.fixture(scope="session")
def grid_all():
    """A Grid with all features"""
    nodes = Nodes(
        id=[0, 1, 2],
        dmax=[1.2, 2.2, 3.3],
        content_pk=[0, 1, 2],
        coordinates=[(1, 1), (2, 2), (3, 3)],
        bounds=[(0, 0, 1, 1), (0, 0, 0, 0), (0, 0, 0, 0)],
        node_type=[NodeType.NODE_2D_OPEN_WATER] + 2 * [NodeType.NODE_1D_NO_STORAGE],
        display_name=[b"foo", "金蟾", None],
    )
    lines = Lines(
        id=[0, 1, 2, 3, 4],
        dpumax=[1.2, 2.2, 3.3, 4.2, 5.1],
        line=[[0, 1], [1, 2], [2, 0], [0, 2], [2, 1]],
        line_geometries=[
            shapely.linestrings([[1, 1], [2, 2]]),
            shapely.linestrings([[1, 1], [2, 2], [3, 3]]),
            None,
            shapely.linestrings([[0, 0], [10, 0]]),
            shapely.linestrings([[0, 10], [0, 0]]),
        ],
        cross_id1=[4, -9999, 3, 3, 4],
        cross_id2=[-9999, -9999, -9999, -9999, 6],
        content_type=[
            -9999,
            -9999,
            -9999,
            ContentType.TYPE_V2_BREACH,
            ContentType.TYPE_V2_EXCHANGE_LINE,
        ],
        content_pk=[-9999, -9999, -9999, 1, 15],
        ds1d_half=[0.5, 1.0, 1.0, 5.0, 8.0],
    )
    pumps = Pumps(
        id=[0, 1, 2],
        capacity=[0.1, 1.2, 2.5],
        line=[[0, 1], [1, 2], [2, 3]],
    )
    meta = GridMeta(
        epsg_code=28992,
        model_name="test-name",
        grid_settings=GridSettings(
            use_2d=True,
            use_1d_flow=True,
            use_2d_flow=True,
            use_0d_inflow=True,
            grid_space=20.0,
            dist_calc_points=25.0,
            kmax=4,
        ),
        tables_settings=TablesSettings(
            table_step_size=0.05,
            frict_coef=0.03,
            frict_coef_type=9,
        ),
    )
    quadtree_stats = QuadtreeStats(
        **{
            "lgrmin": 5,
            "kmax": 2,
            "mmax": np.array([2, 4], dtype=np.int32),
            "nmax": np.array([3, 5], dtype=np.int32),
            "dx": np.array([1.0, 2.0], dtype=np.float64),
            "dxp": 0.5,
            "x0p": 10.0,
            "y0p": 10.0,
        }
    )
    cross_sections = CrossSections(
        id=[0, 1, 2],
        content_pk=[3, 4, 6],
        width_1d=[0.2, 1.5, 3.1],
        count=[4, 2, -9999],
        offset=[0, 4, -9999],
    )
    cross_sections.tables = np.random.random((6, 2))
    nodes_embedded = Nodes(
        id=[0, 1],
        embedded_in=[1, 2],
        dmax=[2.3, 0.2],
    )
    obstacles = Obstacles(
        id=[0, 1],
        the_geom=[
            shapely.linestrings([[1, 1], [2, 2], [4, 4]]),
            shapely.linestrings([[1, 1], [2, 2], [3, 3]]),
        ],
    )
    breaches = PotentialBreaches(
        id=[0, 1],
        content_pk=[1, -9999],
        line_id=[3, 4],
        the_geom=shapely.points([[0, 2], [1, 1]]),
        maximum_breach_depth=[1.2, np.nan],
        levee_material=[1, -9999],
        code=["a", None],
        display_name=["aa", None],
    )
    surfaces = Surfaces(
        id=[0, 1],
        code=[b"1", b"2"],
        display_name=[b"d1", b"d2"],
        function=[b"f1", b"f2"],
        area=[1.0, 2.0],
        centroid_x=[1.0, 2.0],
        centroid_y=[1.0, 2.0],
        dry_weather_flow=[1.2, 1.2],
        nr_of_inhabitants=[1000.0, 2000.0],
        infiltration_flag=[True, False],
        outflow_delay=[1.1, 1.2],
        storage_limit=[10.0, 20.0],
        fb=[1.0, 0.0],
        fe=[1.0, 0.0],
        ka=[1.0, 0.0],
        kh=[1.0, 0.0],
        surface_class=None,
        surface_inclination=None,
        surface_sub_class=None,
    )

    surface_maps = SurfaceMaps(
        id=[1, 2, 3],
        imp=[1, 2, 1],
        fac=[1.0, 0.0, 0.5],
        nxc=[1.1, 2.1, 1.3],
        nyc=[1.3, 2.3, 1.5],
        pk=[1, 2, 3],
        cci=[1, 2, 2],
    )

    return Grid(
        nodes,
        lines,
        pumps,
        cross_sections,
        surfaces,
        surface_maps,
        nodes_embedded,
        obstacles,
        breaches,
        meta,
        quadtree_stats,
    )


@pytest.fixture
def crs_wkt_28992():
    """A current CRS matching EPSG:28992"""
    return CRS(CRS.from_epsg(28992).to_wkt())


@pytest.fixture
def crs_wkt_28992_legacy():
    """A non-current CRS matching the test DEM file"""
    return CRS.from_wkt(
        'PROJCS["Amersfoort / RD New",GEOGCS["Amersfoort",DATUM["Amersfoort",'
        'SPHEROID["Bessel 1841",6377397.155,299.1528128,AUTHORITY["EPSG","7004"]],'
        "TOWGS84[565.2369,50.0087,465.658,-0.406857,0.350733,-1.87035,4.0812],"
        'AUTHORITY["EPSG","6289"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],'
        'UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4289"]],'
        'PROJECTION["Oblique_Stereographic"],PARAMETER["latitude_of_origin",52.1561605555556],'
        'PARAMETER["central_meridian",5.38763888888889],PARAMETER["scale_factor",0.9999079],'
        'PARAMETER["false_easting",155000],PARAMETER["false_northing",463000],'
        'UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["Easting",EAST],AXIS["Northing",NORTH],'
        'AUTHORITY["EPSG","28992"]]'
    )



from pathlib import Path

import typer

from .application import make_gridadmin


def _progress_callback(progress: float, message: str):
    typer.echo("Progress: %d, Message: %s" % (progress * 100, message))


def main(
    sqlite_path: Path,
    dem_path: Path,
    out_path: Path,
    upgrade: bool = False,
):
    make_gridadmin(
        sqlite_path=sqlite_path,
        dem_path=dem_path,
        out_path=out_path,
        progress_callback=_progress_callback,
        upgrade=upgrade,
    )


def run():
    typer.run(main)


if __name__ == "__main__":
    run()
